# Generated by Django 5.1.1 on 2024-09-17 21:28

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0014_alter_file_file"),
    ]

    operations = [
        migrations.RenameField(
            model_name="documenttemplate",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="documenttemplate",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="documenttemplatefreevariable",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="documenttemplatefreevariable",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="documenttemplateheader",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="documenttemplateheader",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="kpi",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="kpi",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="sqlquery",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="sqlquery",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="sqlqueryfilter",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="sqlqueryfilter",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="userkpidashboardviewpreference",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="userkpidashboardviewpreference",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.RenameField(
            model_name="userkpilistviewpreference",
            old_name="created_at",
            new_name="datetime_created",
        ),
        migrations.RenameField(
            model_name="userkpilistviewpreference",
            old_name="updated_at",
            new_name="datetime_updated",
        ),
        migrations.AddField(
            model_name="file",
            name="datetime_created",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="file",
            name="datetime_updated",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 5.1.1 on 2025-01-08 00:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0062_alter_sqlqueryfilter_data_type_todo"),
    ]

    operations = [
        migrations.RenameField(
            model_name="todo",
            old_name="user",
            new_name="requested_for",
        ),
        migrations.AddField(
            model_name="todo",
            name="title",
            field=models.CharField(
                default=1, help_text="The name of the todo", max_length=255
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="todo",
            name="content",
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-20 22:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "test_app",
            "0018_alter_accommodation_options_alter_customer_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="accommodation",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="customer",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                ),
                "verbose_name": "customer",
                "verbose_name_plural": "customers",
            },
        ),
        migrations.AlterModelOptions(
            name="employee",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="employeeonaccommodation",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="equipment",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="function",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "permissions": (
                    (
                        "bulk_delete_%(class)s",
                        "Can delete multiple objects from %(verbose_name)s at once ",
                    ),
                    (
                        "bulk_change_%(class)s",
                        "Can update multiple objects from %(verbose_name)s at once",
                    ),
                )
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-20 23:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0051_alter_applicationfield_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="sqlquery",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                    "export",
                ),
                "managed": True,
                "permissions": [("execute_sql_query", "Can execute SQL queries")],
                "verbose_name": "SQL Query",
                "verbose_name_plural": "SQL Queries",
            },
        ),
    ]



from django.db import models
from django.core.exceptions import ValidationError
from django.urls import reverse
from bloomerp.models.core import BloomerpModel
from django.contrib.contenttypes.models import ContentType
from bloomerp.models.fields import CodeField

# ---------------------------------
# SQL Query Filter Model
# ---------------------------------
from bloomerp.utils.sql import SqlQueryExecutor
class SqlQueryFilter(BloomerpModel):
    '''
    A model to store filters for KPIs. Ech filter has a name and a query.
    '''
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_sql_query_filter'
        verbose_name = 'SQL Query Filter'
        verbose_name_plural = 'SQL Query Filters'

    class SqlQueryFilterOperations(models.TextChoices):
        EQUAL = '='
        GREATER = '>'
        LESS = '<'
        GREATER_OR_EQUAL = '>='
        LESS_OR_EQUAL = '<='
        NOT_EQUAL = '!='
        LIKE = 'LIKE'
        NOT_LIKE = 'NOT LIKE'
        IN = 'IN'
        NOT_IN = 'NOT IN'
        IS = 'IS'
        IS_NOT = 'IS NOT'
        BETWEEN = 'BETWEEN'
        NOT_BETWEEN = 'NOT BETWEEN'
        REGEXP = 'REGEXP'
        NOT_REGEXP = 'NOT REGEXP'
        EXISTS = 'EXISTS'
        NOT_EXISTS = 'NOT EXISTS'

    class SqlQueryFilterDataTypes(models.TextChoices):
        STRING = 'STRING'
        INTEGER = 'INTEGER'
        FLOAT = 'FLOAT'
        DATE = 'DATE'
        DATETIME = 'DATETIME'
        BOOLEAN = 'BOOLEAN'
        MODEL = 'MODEL'
        USER = 'USER'

    name = models.CharField(
        max_length=255,
        help_text='Name of the filter'
        )
    
    column = models.CharField(
        max_length=255, 
        help_text='Column name to filter on')
    
    filter_operation = models.CharField(
        max_length=255, 
        choices=SqlQueryFilterOperations,
        default=SqlQueryFilterOperations.EQUAL,
        help_text='Type of the filter'
        )
    
    data_type = models.CharField(
        max_length=255,
        choices=SqlQueryFilterDataTypes,
        default=SqlQueryFilterDataTypes.STRING,
        help_text='Data type of the filter'
        )

    filter_options = models.JSONField(blank=True, null=True)

    # Query executor
    executor = SqlQueryExecutor()

    def __str__(self):
        return self.name
    
    

# ---------------------------------
# SQL Query Model
# ---------------------------------
from django.contrib.auth.hashers import make_password
import time
class SqlQuery(BloomerpModel):
    '''
    A model to store SQL queries that can be used in Widgets
    '''
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'sql_query'
        verbose_name = 'SQL Query'
        verbose_name_plural = 'SQL Queries'
        permissions = [
            ('execute_sql_query', 'Can execute SQL queries')
            # Maybe add more permissions here corresponding to the actions that can be performed on the query
        ]

    name = models.CharField(max_length=255)
    query = CodeField(language='sql', help_text='SQL Query to execute')
    filters = models.ManyToManyField(SqlQueryFilter, blank=True)

    # String fields
    search_fields = ['name']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        start = time.time()

        # Initialize the query executor
        if self.pk is not None:
            self.executor = SqlQueryExecutor(
                cache_time=60,
                cache_id= str(self.pk) + self.name 
            )
        else:
            self.executor = SqlQueryExecutor()
    
    def is_safe(self):
        '''
        Check if the query is a SELECT statement
        '''
        return self.executor.is_safe(self.query)

    def result_dataframe(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_df(self.query)
        
    def result_dict(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_dict(self.query)
    
    def result_raw(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_raw(self.query)
    
    def result_value(self):
        '''
        Execute the query and return the result
        '''
        return self.executor.execute_to_first_value(self.query)

    def clean(self) -> None:
        '''
        Clean method does the following:
        - Check if the query is safe
        - Check if the query actually returns a result
        '''
        errors = {}

        # Check if the query is safe
        if not self.is_safe():
            errors['query'] = 'Unsafe query'

        # Check if the query returns a result
        result = self.executor.is_valid(self.query)
        if result != True:
            errors['query'] = result

        if errors:
            raise ValidationError(errors)


    def __str__(self):
        return self.name

    
# ---------------------------------
# Widget Model
# ---------------------------------
from pandas import DataFrame
class Widget(BloomerpModel):
    OUTPUT_TYPE_CHOICES = [
        ('table','Table'),
        ('value','Value'),
        ('line','Line chart'),
        ('bar','Bar chart'),
        ('pie','Pie chart'),
        ('scatter','Scatter chart'),
        ('histogram','Histogram')
    ]

    VALUE_AGGREGATE_CHOICES = [
        ('sum', 'Sum'),
        ('average', 'Average'),
        ('count', 'Count'),
        ('min', 'Minimum'),
        ('max', 'Maximum'),
        ('first', 'First value')
    ]

    name = models.CharField(max_length=255) # Name of the KPI
    query = models.ForeignKey(SqlQuery, on_delete=models.CASCADE) # Query to execute
    description = models.TextField(blank=True,null=True) # Description of the KPI
    output_type = models.CharField(choices=OUTPUT_TYPE_CHOICES, max_length=10)
    options = models.JSONField(blank=True,null=True)


    string_search_fields = ['name', 'description']
    allow_string_search = True


    def __str__(self):
        return self.name
    
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_widget'
    
    def clean(self):
        # SAFETY CHECKS
        if not self.query.is_safe:
            raise ValidationError('Unsafe query')

        # Get columns
        dataframe = self.query.result_dataframe()
        columns = dataframe.columns.tolist()

        # TABLE SPECIFIC VALIDATION
        if self.output_type == 'table':
            if self.options is None:
                raise ValidationError('Options are required for table KPIs')

            if 'columns' not in self.options:
                raise ValidationError('Columns are required for table KPIs')
            
            for col in self.options['columns']:
                if col not in columns:
                    raise ValidationError(f'Column {col} does not exist in the query result')
            if 'limit' not in self.options:
                self.options['limit'] = 10

            

        # GRAPH SPECIFIC VALIDATION
        if (self.output_type == 'bar' or self.output_type == 'line' or self.output_type == 'scatter') and self.options is None:
            raise ValidationError('Options are required for chart KPIs')

        
        # BAR / LINE / SCATTER SPECIFIC VALIDATION
        if self.output_type == 'bar' or self.output_type == 'line' or self.output_type == 'scatter':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for bar chart KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
            
            if 'y' not in self.options:
                raise ValidationError('Y-axis is required for bar chart KPIs')
            else:
                if self.options['y'] not in columns:
                    raise ValidationError(f'Column {self.options["y"]} does not exist in the query result')

            if 'group_by' in self.options:
                if self.options['group_by'] not in columns:
                    raise ValidationError(f'Column {self.options["group_by"]} does not exist in the query result')

        # PIE SPECIFIC VALIDATION     
        if self.output_type == 'pie':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for pie chart KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
            
            if 'group_by' not in self.options:
                raise ValidationError('Groupby column is required for pie chart KPIs')
            else:
                if self.options['group_by'] not in columns:
                    raise ValidationError(f'Column {self.options["y"]} does not exist in the query result')
                
        # HISTOGRAM SPECIFIC VALIDATION
        if self.output_type == 'histogram':
            if 'x' not in self.options:
                raise ValidationError('X-axis is required for histogram KPIs')
            else:
                if self.options['x'] not in columns:
                    raise ValidationError(f'Column {self.options["x"]} does not exist in the query result')
                
        # VALUE SPECIFIC VALIDATION
        if self.output_type == 'value':
            if 'color' not in self.options:
                self.options['color'] = 'gray'
            
            if 'prefix' not in self.options:
                self.options['prefix'] = ''

            if 'suffix' not in self.options:
                self.options['suffix'] = ''
            
            if 'icon' not in self.options:
                self.options['icon'] = 'bi bi-bar-chart'

            if 'column' not in self.options:
                raise ValidationError('Column is required for value KPIs')

            if 'aggregate' not in self.options:
                self.options['aggregate'] = 'first'
            
            if self.options['aggregate'] not in [agg[0] for agg in self.VALUE_AGGREGATE_CHOICES]:
                raise ValidationError('Invalid aggregate type')
            
    @property
    def result_value(self):
        '''
        Execute the query and return the result
        '''
        dataframe:DataFrame = self.query.result_dataframe()

        if self.output_type == 'value':
            column = self.options['column']
            aggregate = self.options['aggregate']

            try:
                if aggregate == 'sum':
                    return dataframe[column].sum()
                elif aggregate == 'average':
                    return dataframe[column].mean()
                elif aggregate == 'count':
                    return dataframe[column].count()
                elif aggregate == 'min':
                    return dataframe[column].min()
                elif aggregate == 'max':
                    return dataframe[column].max()
                elif aggregate == 'first':
                    return dataframe[column].iloc[0]
            except:
                self.options['aggregate'] = 'first'
                self.save()
                return dataframe[column].iloc[0]
        else:
            return 'Value result not available for this KPI type'

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)




from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models import File, FileFolder, User
from django.contrib.contenttypes.models import ContentType
from bloomerp.forms.core import BloomerpModelForm
from django.forms.models import modelform_factory
from bloomerp.utils.filters import dynamic_filterset_factory
from django.contrib.auth.decorators import login_required

@login_required
@route('files')
def files(request: HttpRequest) -> HttpResponse:
    '''
    Component for rendering a list of files and folders.
    The files can be filtered by content type, object id, and folder.
    
    Variables:
    - content_type_id: The content type id related to the file.
    - object_id: The object id related to the file.
    - query: A search query to filter files by name.
    - folder_id: The folder id to filter files by folder.
    - sort: The field to sort files by.

    '''
    # Some logic for permissions here
    if not request.user.has_perm('bloomerp.view_file'):
        return HttpResponse('User does not have permission to view files')

    # get the accessible content types for the user
    user : User = request.user
    accessible_content_types = user.get_content_types_for_user('view')


    # initialize variables
    content_type_id = request.GET.get('content_type_id', None)
    object_id = request.GET.get('object_id', None)
    query = request.GET.get('query', None)
    folder_id = request.GET.get('folder_id', None)
    sort = request.GET.get('sort', None)
    target = request.htmx.target


    # Convert empty strings to None
    if content_type_id in ['', 'None']:
        content_type_id = None
    if object_id in ['', 'None']:
        object_id = None
    if folder_id in ['', 'None']:
        folder_id = None
    if sort == '':
        sort = None

    FilterSet = dynamic_filterset_factory(File)

    # Initialize create folder form
    create_folder_form = modelform_factory(FileFolder, form=BloomerpModelForm, fields=['name', 'content_types'])(model=FileFolder, user=request.user, initial={'content_types': [content_type_id] if content_type_id else None})

    # First, we check if the user is trying to filter by folder
    if folder_id:
        try:
            folder = FileFolder.objects.get(id=folder_id)
            # If the user is filtering by folder, we only show files in that folder
            files = folder.files.all()

            # Filter files by content type and object id if they are provided
            if object_id:
                files = files.filter(object_id=object_id)

            # Filter files by query if it exists
            if query:
                files = files.filter(name__icontains=query)

            # Get all folders within the current folder
            folders = FileFolder.objects.filter(parent=folder).order_by('name')

            # Create update folder form
            update_folder_form = modelform_factory(FileFolder, form=BloomerpModelForm, fields=['name', 'content_types'])(model=FileFolder, user=request.user, instance=folder)

            if content_type_id:
                folders = folders.filter(content_types__in=[content_type_id])

        except FileFolder.DoesNotExist:
            # Handle the case where the folder does not exist
            folders = FileFolder.objects.none()  # or some other error handling logic
    else:
        # If no folder_id is specified, fetch root-level folders (parent=None)
        folders = FileFolder.objects.filter(parent=None)

        if content_type_id:
            folders = folders.filter(content_types__in=[content_type_id])
            

        update_folder_form = None

        # If the user is not filtering by folder, we show all files, optionally filtering by query
        if query:
            files = File.objects.filter(name__icontains=query)
        else:
            files = File.objects.all()

        # Filter files by content type and object id if they are provided
        if content_type_id:
            files = files.filter(content_type_id=content_type_id)

        if object_id:
            files = files.filter(object_id=object_id)

        # Exclude files that are in a folder
        files = files.filter(folders=None)

    # Sort files by name if sort is not specified
    if not sort:
        files = files.order_by('name')
    elif sort:
        try:
        # Sort files by name in ascending order
            files = files.order_by(sort)
            folders = folders.order_by(sort)
        except ValueError:
            # Handle the case where the sort parameter is invalid
            pass
    
    # Add the current url to the context
    url = request.path
    
    if folder_id or content_type_id or object_id:
        url += '?'  # Add a ? to the url if there are any query parameters
    if folder_id:
        url += f'folder_id={folder_id}'
    if content_type_id:
        url += f'&content_type_id={content_type_id}'
    if object_id:
        url += f'&object_id={object_id}'
    
    # Filter files by content type and object id if they are provided
    filterset = FilterSet(request.GET, queryset=files) # Initialize files as empty queryset in case it's not used later
    files = filterset.qs

    # Filter on accessible content types, including files that are not mapped to a content type
    files = files.filter(content_type_id__in=accessible_content_types) | files.filter(content_type_id__isnull=True)

    context = {
        'files': files,
        'folders': folders,
        'current_folder': folder if folder_id else None,
        'sort': sort,
        'content_type_id': content_type_id,
        'object_id': object_id,
        'file_content_type_id': ContentType.objects.get_for_model(File).id,
        'url': url,
        'update_folder_form': update_folder_form,
        'create_folder_form': create_folder_form,
        'target': target,
        }
    return render(request, 'components/files.html', context)



# Generated by Django 5.1.1 on 2025-01-11 21:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0070_alter_comment_object_id_alter_file_object_id_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="aiconversation",
            name="auto_named",
            field=models.BooleanField(
                default=False, help_text="Whether the conversation has been auto-named"
            ),
        ),
    ]




BLOOMERP_APPS = [
    "bloomerp",
    "django_htmx",
    "crispy_forms",
    "crispy_bootstrap5",
    "rest_framework",
    "django_filters",
]

BLOOMERP_MIDDLEWARE = [
    "bloomerp.middleware.HTMXPermissionDeniedMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
]

BLOOMERP_USER_MODEL = "bloomerp.User"




# Generated by Django 5.1.1 on 2024-12-16 13:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0057_alter_documenttemplate_footer"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="documenttemplate",
            name="model_variable",
            field=models.ForeignKey(
                blank=True,
                help_text="Model variable of the document template. Can be used to parse objects from the model into the template.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models.fields import TextEditorField
from bloomerp.widgets.foreign_key_widget import ForeignKeyWidget
from bloomerp.widgets.multiple_model_select_widget import MultipleModelSelect
from bloomerp.widgets.text_editor import RichTextEditorWidget
from django.forms.fields import CharField
from django.forms.widgets import Textarea
from bloomerp.models import Link, Widget
import uuid
from django.contrib.auth.decorators import login_required


@login_required
@route('create_workspace_item')
def create_workspace_item(request: HttpRequest) -> HttpResponse:
    # Some permissions check (can user create dashboard item?)
    if not request.user.has_perm('bloomerp.view_workspace'):
        return HttpResponse('Permission denied', status=403)

    # Get the action (get_input, get_snippet)
    action = request.GET.get('action', 'get_input_field')
    if action == 'get_input_field':
        # Get initial data
        item_type = request.GET.get('item_type', 'text')
        if item_type == 'text':
            widget = Textarea()
            field_html = widget.render(name='text', value='', attrs={'class': 'form-control'})
            return HttpResponse(field_html)
        elif item_type == 'header':
            field = CharField()
            field_html = field.widget.render(name='header', value='', attrs={'class': 'form-control'})
            return HttpResponse(field_html)
        elif item_type == 'link':
            field = ForeignKeyWidget(model=Link)
            field_html = field.render(name='link', value=None, attrs={})
            return HttpResponse(field_html)
        elif item_type == 'link_list':
            field = MultipleModelSelect(model=Link)
            field_html = field.render(name='link_list', value=None, attrs={})
            return HttpResponse(field_html)
        elif item_type == 'widget':
            field = ForeignKeyWidget(model=Widget)
            field_html = field.render(name='widget', value=None, attrs={})
            return HttpResponse(field_html)
    elif action == 'get_snippet':
        item = {}
        item['data'] = {}
        item['size'] = 12
        if request.GET.get('text'):
            item['data']['text'] = request.GET.get('text')
            item['type'] = 'text'
        elif request.GET.get('header'):
            item['data']['text'] = request.GET.get('header')
            item['type'] = 'header'
        elif request.GET.get('link'):
            item['data']['link_id'] = request.GET.get('link')
            item['type'] = 'link'
        elif request.GET.get('link_list'):
            item['data']['links'] = request.GET.getlist('link_list')
            item['type'] = 'link_list'
        elif request.GET.get('widget'):
            item['data']['widget_id'] = request.GET.get('widget')
            item['type'] = 'widget'

        item['id'] = uuid.uuid4()
        return render(request, 'snippets/workspace_item.html', {'item': item})
            


# Generated by Django 5.1.1 on 2024-09-23 21:01

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0017_bookmark"),
    ]

    operations = [
        migrations.CreateModel(
            name="DocumentTemplateStyling",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                ("styling", models.TextField(default="")),
            ],
            options={
                "db_table": "bloomerp_document_template_styling",
                "managed": True,
            },
        ),
        migrations.AddField(
            model_name="documenttemplate",
            name="styling",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="bloomerp.documenttemplatestyling",
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-15 22:11

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0002_alter_customer_options_alter_order_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="order",
            options={
                "permissions": [
                    (
                        "bulk_delete_%(class)",
                        "Can delete multiple objects from %(class) at once",
                    ),
                    (
                        "bulk_change_%(class)",
                        "Can update multiple objects from %(class) at once",
                    ),
                ]
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "permissions": [
                    (
                        "bulk_delete_%(class)",
                        "Can delete multiple objects from %(class) at once",
                    ),
                    (
                        "bulk_change_%(class)",
                        "Can update multiple objects from %(class) at once",
                    ),
                ]
            },
        ),
    ]



from django import forms
from bloomerp.models import DocumentTemplate, DocumentTemplateFreeVariable
from django.apps import apps
from bloomerp.widgets.foreign_key_widget import ForeignKeyWidget

# ---------------------------------
# Free Variable Form
# ---------------------------------
class FreeVariableForm(forms.Form):
    def __init__(self, document_template: DocumentTemplate, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set document template
        self.document_template = document_template

        # Add fields for free variables
        free_variables = document_template.free_variables.all()
        for variable in free_variables:
            variable : DocumentTemplateFreeVariable
            field_kwargs = {
                'label': f'{variable.name} (variable)',
                'required': variable.required,
                'help_text': variable.help_text,
            }
            if variable.variable_type == 'date':
                field_kwargs['widget'] = forms.DateInput(
                    attrs={'type': 'date', 'class': 'form-control'})
                field = forms.DateField(**field_kwargs)
            elif variable.variable_type == 'boolean':
                field = forms.BooleanField(**field_kwargs)

            elif variable.variable_type == 'integer':
                field = forms.IntegerField(**field_kwargs)

            elif variable.variable_type == 'float':
                field = forms.FloatField(**field_kwargs)

            elif variable.variable_type == 'text':
                field = forms.CharField(**field_kwargs)
            elif variable.variable_type == 'list':
                # Create options from choices
                options = variable.options.split(',')
                field_kwargs['choices'] = [(option, option)
                                           for option in options]

                field = forms.ChoiceField(**field_kwargs)
            elif variable.variable_type == 'model':
                # Retrieve the model
                try:
                    model = variable.options.split(',')
                    # Get the model class using the apps module without specifying app_label
                    model_class = apps.get_model(
                        app_label=model[0], model_name=model[1])

                    # Get a queryset for the model
                    queryset = model_class.objects.all()

                    # create the field
                    field = forms.ModelChoiceField(queryset=queryset)
                except:
                    # Handle the case where the model is not found
                    pass

            else:
                field = forms.CharField(**field_kwargs)

            self.fields[variable.slug] = field

        # Add hidden field for the document template
        self.fields['document_template_id'] = forms.CharField(widget=forms.HiddenInput(attrs={'value': document_template.pk}))

# ---------------------------------
# Generate Document Form
# ---------------------------------
class GenerateDocumentForm(FreeVariableForm):
    '''
    Form that is used to generate a document from a document template.
    It is not specific to any object, and can therefore be used outside of the detail-view.
    '''
    instance = None
    persist = False

    # Add a field for persisting the document
    save_file = forms.BooleanField(
        label='Save document',
        required=False,
        help_text='Check this box to save the document to the database')
    

    def __init__(
            self, 
            document_template: DocumentTemplate,
            add_persist_field : bool = True,
            *args, **kwargs):
        super().__init__(document_template, *args, **kwargs)

        if document_template.model_variable:
            related_model = document_template.model_variable.model_class()
            field = forms.ModelChoiceField(
                queryset=related_model.objects.all(), widget=ForeignKeyWidget(related_model))
            self.fields['generate_document_for'] = field

            self.fields['generate_document_for'].label = f'Generate document for {related_model._meta.verbose_name}'

            # Set position of the field to the top
            self.order_fields(['save_file','generate_document_for'])

        if not add_persist_field:
            self.fields.pop('save_file')

    def clean(self):
        cleaned_data = super().clean()

        # Check if the generate_document_for field is present
        if 'generate_document_for' in cleaned_data:
            self.instance = cleaned_data['generate_document_for']    

        # Check if the save_file field is present
        if 'save_file' in cleaned_data:
            self.persist = cleaned_data['save_file']

        return cleaned_data
    
    def has_persist_field(self) -> bool:
        '''Returns whether the form has a persist field or not.'''
        if self.fields.get('save_file',False):
            return True
        else:
            return False
    
    


# Generated by Django 5.1.1 on 2024-10-02 23:06

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0026_filefolder_parent"),
    ]

    operations = [
        migrations.AddField(
            model_name="filefolder",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="filefolder",
            name="datetime_created",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="filefolder",
            name="datetime_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="filefolder",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.forms.widgets import Widget, ClearableFileInput
from bloomerp.models.core import File
import uuid


class BloomerpFileFieldWidget(Widget):
    template_name = 'widgets/bloomerp_file_field_widget.html'
    file_input = None

    def __init__(self, attrs=None):
        self.file_input = ClearableFileInput(attrs=attrs)
        super().__init__(attrs)

    def get_context(self, name, value, attrs):
        # Set default attributes if they are not provided
        if attrs is None:
            attrs = {}

        # Check if it was marked as invalid, in that case delete the file
        if attrs.get('aria-invalid', 'false') == 'true':
            invalid = True
            File.objects.filter(pk=value).delete()
            value = None
        else:
            invalid = False
        
        # Add your custom attributes, like 'class' for Bootstrap form control
        attrs.setdefault('class', 'form-control')

        # Set value for the file field if it exists
        # This will be used to display the file name in the template
        if value:
            file_obj = File.objects.get(pk=value)
            value = file_obj.file

        # Use ClearableFileInput's get_context but with modified attrs
        context = self.file_input.get_context(name, value, attrs)

        if invalid:
            context['invalid'] = True

        if value:
            context['current_file'] = file_obj

        return context

    def value_from_datadict(self, data, files, name):
        file = self.file_input.value_from_datadict(data, files, name)

        # Check if there already is a current file instance
        current_file = data.get(name + '_current', None)
    
        # If the file is cleared, the file instance will be False, so we should return None
        if file is False:
            return None

        # If no file is provided, the value_from_datadict method will return None
        # However, we need to check if there is a current file instance
        if file is None and not current_file:
            # Case when there is no current file and no new file
            return None
        elif file is None and current_file:
            # Case when there is a current file but no new file
            return current_file
        elif file is not None and not current_file:
            # Create or update the file instance
            file_obj = File(
                file=file,
                name=file.name,
                persisted=False
            )
            # Temporarily save the file in memory
            file_obj.save()
            return file_obj.pk
        else:
            # Case when there is a current file and a new file
            # Update the current file instance
            file_obj = File.objects.get(pk=current_file)
            file_obj.file = file
            file_obj.name = file.name
            file_obj.save()
            return file_obj.pk
        

    # Pass-through methods to ClearableFileInput (you can omit these if already implemented)
    def format_value(self, value):
        return self.file_input.format_value(value)

    def clear_checkbox_name(self, name):
        return self.file_input.clear_checkbox_name(name)

    def clear_checkbox_id(self, name):
        return self.file_input.clear_checkbox_id(name)

    def is_initial(self, value):
        return self.file_input.is_initial(value)

    def value_omitted_from_data(self, data, files, name):
        return self.file_input.value_omitted_from_data(data, files, name)

    def use_required_attribute(self, initial):
        return self.file_input.use_required_attribute(initial)



from django.db.models.signals import post_save, post_delete, pre_save
from django.dispatch import receiver
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

def notification_handler(
        sender, 
        instance : Model, 
        created: bool,
        **kwargs):
    if created:
        # Do something
        print('New notification created!', instance, sender)
    else:
        # Do something
        print('Notification updated!', instance, sender)

# Connect all models to the notification handler



from django.apps import apps
from django.shortcuts import render
from django.http import HttpRequest, HttpResponse
from bloomerp.models import BloomerpModel, User
from bloomerp.utils.router import route
from bloomerp.models import Link
from bloomerp.utils.models import search_content_types_by_query, search_objects_by_content_types
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required

@login_required
@route('search_results')
def search_results(request: HttpRequest) -> HttpResponse:
    '''
    Component that returns search results for a given query.
    This component is used to display search results in the search bar dropdown.
    '''
    query = request.GET.get('search_results_query', '')  # Retrieve the search query from the GET parameters
    results = []
    links = None
    not_found_response = HttpResponse('<li class="dropdown-item">No results found</li>')


    if query == '':
        return not_found_response

    try:
        if query.startswith('/') and query[1] != '/':
            # In this case we want to return links for a specific model
            # Name is optional and comes after a : in the query
            sub_query = query[1:]

            name = None
            if ':' in sub_query:
                sub_query, name = sub_query.split(':')

            if '/' in sub_query:
                content_type_query = sub_query.split('/')[0]
            else:
                content_type_query = sub_query


            content_types = search_content_types_by_query(content_type_query)
            list_level_links = []

            if '/' in sub_query:
                object_query = sub_query.split('/')[-1]

                results = search_objects_by_content_types(object_query, content_types=content_types, limit=5, user=request.user)

                context = {
                    "results": results,
                    "query": query
                }

                return render(request, 'components/search_results.html', context)
                            
            else:
                for content_type in content_types:
                    links = Link.objects.filter(content_type=content_type, level='LIST')
                    if name:
                        links = links.filter(name__icontains=name)
                    
                    # If there are links for the model, add them to the list, otherwise skip
                    if links.exists():
                        list_level_links.append({
                            'model_name': content_type.model_class()._meta.verbose_name,
                            'links': links
                        })
                    
                return render(request, 'components/search_results.html', {'list_level_links': list_level_links})
        
        if query.startswith('//'):
            # In this case, we want to return app level links
            sub_query = query[2:]
            app_level_links = Link.objects.filter(level='APP', name__icontains=sub_query)
            return render(request, 'components/search_results.html', {'app_level_links': app_level_links})

        else:
            # Get all models from the app
            content_types = User.get_content_types_for_user(request.user)
            
            results = search_objects_by_content_types(query, content_types=content_types, limit=5, user=request.user)

            context = {
                "results": results,
                "query": query
            }

            return render(request, 'components/search_results.html', context)
    except Exception as e:
        print(e)
        return not_found_response
    




from django import forms

# ---------------------------------
# User Detail View Preference Form
# ---------------------------------
from bloomerp.models import UserDetailViewPreference

class UserDetailViewPreferenceForm(forms.Form):
    field = forms.CharField(max_length=100, required=False)
    id = forms.IntegerField(required=False)
    position = forms.ChoiceField(required=False, choices=UserDetailViewPreference.POSITION_CHOICES, initial=UserDetailViewPreference.POSITION_CHOICES[0][0])
    is_used = forms.BooleanField(required=False)

class UserDetailViewPreferenceFormset(forms.BaseFormSet):
    user = None
    def save(self, *args, **kwargs):
        for form in self.forms:
            form: UserDetailViewPreferenceForm
            if form.cleaned_data.get('id'):
                # Extract relevant data from the form
                id = form.cleaned_data.get('id',None)
                is_used = form.cleaned_data.get('is_used',False)
                position = form.cleaned_data.get('position',False)

                # Check if UserListViewPreference with specified conditions exists
                existing_preference = UserDetailViewPreference.objects.filter(
                    user=self.user,
                    application_field_id=id,
                ).first()

                # Implement your business logic here
                if existing_preference:
                    # UserListViewPreference with given conditions exists
                    if not is_used:
                        # If is_used is False, delete the existing preference
                        existing_preference.delete()
                    # Note: If is_used is True, no further action is required as we want to create it if it doesn't exist
                else:
                    # UserListViewPreference with given conditions does not exist
                    if is_used:
                        
                        # If is_used is True, create a new preference
                        UserDetailViewPreference.objects.create(
                            user=self.user,
                            application_field_id=id,
                            position=position
                        )


# ---------------------------------
# User selection form
# ---------------------------------
from bloomerp.models import User
class UserSelectionForm(forms.Form):
    users = forms.ModelChoiceField(
        queryset=User.objects.all(),
        empty_label="Select a user",  # Optional: Display a default label
        required=False
    )


# ---------------------------------
# User Creation Form
# ---------------------------------
from django.contrib.auth.forms import UserCreationForm
class BloomerpUserCreationForm(UserCreationForm):
    class Meta(UserCreationForm.Meta):
        model = User
        fields = UserCreationForm.Meta.fields + ('email',)




# Generated by Django 5.1.1 on 2024-10-20 20:25

import bloomerp.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0036_alter_link_level_userdetailviewtabs"),
    ]

    operations = [
        migrations.AlterField(
            model_name="documenttemplatestyling",
            name="styling",
            field=bloomerp.models.fields.CodeField(default="", language="css"),
        ),
        migrations.AlterField(
            model_name="user",
            name="avatar",
            field=models.ImageField(
                blank=True, help_text="The user's avatar", null=True, upload_to="users/"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="date_view_preference",
            field=models.CharField(
                choices=[
                    ("d-m-Y", "Day-Month-Year (15-08-2000)"),
                    ("m-d-Y", "Month-Day-Year (08-15-2000)"),
                    ("Y-m-d", "Year-Month-Day (2000-08-15)"),
                ],
                default="d-m-Y",
                help_text="The date format to be used in the application",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="datetime_view_preference",
            field=models.CharField(
                choices=[
                    ("d-m-Y H:i", "Day-Month-Year Hour:Minute (15-08-2000 12:30)"),
                    ("m-d-Y H:i", "Month-Day-Year Hour:Minute (08-15-2000 12:30)"),
                    ("Y-m-d H:i", "Year-Month-Day Hour:Minute (2000-08-15 12:30)"),
                ],
                default="d-m-Y H:i",
                help_text="The datetime format to be used in the application",
                max_length=20,
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest, StreamingHttpResponse
from bloomerp.utils.llm import BloomerpOpenAI, BloomerpLangChain
from bloomerp.models import ApplicationField, DocumentTemplate, AIConversation
from django.contrib.auth.decorators import login_required
from django.conf import settings
import json
from django.core.cache import cache
from bloomerp.langchain_tools import BLOOMAI_TOOLS
import uuid
from django.contrib.auth.models import User
from django.views.decorators.http import require_POST


@require_POST
@login_required
@route('llm_executor')
def llm_executor(request:HttpRequest) -> HttpResponse:
    '''
    Component to execute LLM queries.
    '''
    LLM_QUERY_TYPES = ['sql', 'document_template', 'tiny_mce_content', 'bloom_ai','code']
    

    # Get the json data from the request
    json_data : dict = json.loads(request.body)

    query_type = json_data.get('query_type', None)
    query = json_data.get('query', None)
    conversation_id = json_data.get('conversation_id', None)

    # Some preprocessing
    if not query_type:
        return HttpResponse('No llm query type provided')
    
    # Check if the query type is valid
    if query_type not in LLM_QUERY_TYPES:
        return HttpResponse('Invalid llm query type, must be one of: ' + ', '.join(LLM_QUERY_TYPES))
    
    # Check if the query is provided
    openai_key = settings.BLOOMERP_SETTINGS.get('OPENAI_API_KEY', None)
    if not openai_key:
        return HttpResponse('OpenAI key not found in settings')
    
    # Init the OpenAI class
    openai = BloomerpOpenAI(openai_key)

    # Check if the key is valid
    if not openai.is_valid_key():
        return HttpResponse('Invalid OpenAI key')
    
    # Parse uuid
    try:
        id = uuid.UUID(conversation_id)
    except:
        return HttpResponse('Invalid conversation id')

    # Get or create the AIConversation object
    ai_conversation_object, CREATED = AIConversation.objects.get_or_create(
            id=id,
            defaults={
                'conversation_type': query_type,
                'user' : request.user
            }            
        )

    # Get the conversation history
    if CREATED:
        conversation_history = []
    else:
        conversation_history = ai_conversation_object.conversation_history
    
    # Init the BloomerpLangChain executor
    executor = BloomerpLangChain(
        api_key=openai_key, 
        conversation_history=conversation_history
    )   

    return StreamingHttpResponse(
            stream_response(
                executor=executor,
                ai_conversation_object=ai_conversation_object,
                query=query,
                query_type=query_type,
                user=request.user,
                json_data=json_data
            ), 
            content_type='text/html')

def stream_response(
        executor:BloomerpLangChain,
        ai_conversation_object:AIConversation,
        query:str,
        user:User,
        query_type:str = 'bloom_ai',
        json_data:dict = None
    ):
    if query_type == 'bloom_ai':
        for item in executor.invoke_bloom_ai(
            query, 
            BLOOMAI_TOOLS, 
            user=user
            ):
            yield item

    elif query_type == 'sql':
        db_tables_and_columns = ApplicationField.get_db_tables_and_columns()
        for item in executor.invoke_sql_query(query, db_tables_and_columns):
            yield item

    elif query_type == 'document_template':
        # Get template id from args
        template_id = json_data['args'].get('template_id', None)

        if not template_id:
            return HttpResponse('Template id not provided')
        
        # Get the document template
        template = DocumentTemplate.objects.get(pk=template_id)
        variables = template.get_variables()

        for item in executor.invoke_document_template(
            query,
            variables
            ):
            yield item

    elif query_type == 'tiny_mce_content':
        for item in executor.invoke_tiny_mce_content(query):
            yield item

    elif query_type == 'code':
        for item in executor.invoke_code(query):
            yield item

    # Get the conversation history
    conversation_history = executor.serialize_conversation_history()

    # Save the conversation history to the object
    ai_conversation_object.conversation_history = conversation_history

    # Autoname the conversation object
    if not ai_conversation_object.auto_named and ai_conversation_object.number_of_messages > 5:
        ai_conversation_object.title = executor.auto_name()
        ai_conversation_object.auto_named = True

    ai_conversation_object.save()
        
    
    




# Generated by Django 5.1.1 on 2024-09-16 22:38

import bloomerp.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "bloomerp",
            "0009_documenttemplatefreevariable_documenttemplateheader_and_more",
        ),
    ]

    operations = [
        migrations.RemoveField(
            model_name="file",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="file",
            name="updated_at",
        ),
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-30 23:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0039_userdetailviewtab"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="userdetailviewtab",
            unique_together={("user", "link")},
        ),
        migrations.RemoveField(
            model_name="userdetailviewtab",
            name="content_type",
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-22 17:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "test_app",
            "0024_accommodation_avatar_customer_avatar_employee_avatar_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="employee",
            name="accommodations",
            field=models.ManyToManyField(
                related_name="employees",
                through="test_app.EmployeeOnAccommodation",
                to="test_app.accommodation",
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.core.files import File
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from bloomerp.utils.model_io import BloomerpModelIO
from django.forms.models import modelformset_factory
from django.forms import BaseFormSet
from django.views.decorators.http import require_POST


@require_POST
@login_required
@route('bulk_upload_table')
def bulk_upload_table(request:HttpRequest) -> HttpResponse:
    '''
    Component that allows for bulk uploading of data into a model.

    POST Parameters:
    - bulk_upload_file: The file to upload
    - bulk_upload_content_type_id: The content type id of the model to upload to

    '''
    bulk_upload_file:File = request.FILES.get('bulk_upload_file')
    content_type_id = request.POST.get('bulk_upload_content_type_id')
    
    print(request.POST)

    # Get the model from the content type id
    if not content_type_id:
        return HttpResponse('content_type_id required in the query parameters', status=400)
    else:
        try:
            content_type_id = int(content_type_id)
            model = ContentType.objects.get(id=content_type_id).model_class()
        except ValueError:
            return HttpResponse('Invalid content type id', status=400)
        except ContentType.DoesNotExist:
            return HttpResponse('Invalid content type id', status=400)

    # Permissions check
    if not request.user.has_perm(f'{model._meta.app_label}.bulk_add_{model._meta.model_name}'):
        return HttpResponse('Permission denied', status=403)

    context = {}
    model_io = BloomerpModelIO(model)

    if bulk_upload_file:
        # Flow in case of a file upload
        try:
            data = model_io.import_from_template(bulk_upload_file)
        except Exception as e:
            return HttpResponse(f'Error importing file: {e}', status=200)
    
        fields = data[0].keys()
        Formset : BaseFormSet = modelformset_factory(model, fields=fields, extra=len(data)) 

        # Create a blank formset
        formset = Formset(queryset=model.objects.none(), initial=data, prefix = 'bulk_upload')   
    else:
        # Flow in case of a form submission
        # Get the fields that were given based on the form submission
        fields = request.POST.getlist('bulk_upload_fields')

        Formset  = modelformset_factory(model, fields=fields)
        formset : BaseFormSet = Formset(request.POST, request.FILES, prefix = 'bulk_upload')

        if formset.is_valid():
            instances = formset.save(commit=False)

            for form, instance in zip(formset.forms, instances):
                if hasattr(instance, 'created_by'):
                    instance.created_by = request.user
                if hasattr(instance, 'updated_by'):
                    instance.updated_by = request.user
                
                instance.save()

                # Call save_m2m on each form after saving the instance
                if hasattr(form, 'save_m2m'):
                    form.save_m2m()
            
            return HttpResponse('Data saved successfully', status=200)


    context['formset'] = formset
    context['fields'] = fields
    context['content_type_id'] = content_type_id
    return render(request, 'components/bulk_upload_table.html', context)


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.contrib.auth.decorators import login_required
from bloomerp.models import AIConversation
import uuid
from bloomerp.utils.requests import parse_bool_parameter
from django.utils import timezone

@login_required
@route('ai_conversations')
def ai_conversations(request:HttpRequest) -> HttpResponse:
    '''Component that renders ai conversation or a single conversation.

    Args:
        request (HttpRequest): The request object

    GET Parameters:
        target (str): The target to render
        conversation_id (str): The conversation id
        conversation_type (str): The conversation type
        new_conversation (bool): Create a new conversation
        delete (bool): Delete the conversation
        rename (str): Rename the conversation
        args (str): The arguments to be passed to the llm executor
        latest (bool): Get the latest conversation

    Returns:
        HttpResponse: The response object
    '''
    target = request.GET.get('target', None)
    conversation_id = request.GET.get('conversation_id', None)
    conversation_type = request.GET.get('conversation_type', None)
    new_conversation = parse_bool_parameter(request.GET.get('new_conversation'), False)
    delete = parse_bool_parameter(request.GET.get('delete'), False)
    rename = request.GET.get('rename', None)
    args = request.GET.get('args', None)
    latest = parse_bool_parameter(request.GET.get('latest'), False)


    if latest:
        ai_conversations = AIConversation.objects.filter(
            user = request.user,
            conversation_type=conversation_type,
            datetime_updated__gte=timezone.now() - timezone.timedelta(minutes=30)
        ).order_by('-datetime_updated')

        if ai_conversations.exists():
            ai_conversation_object = ai_conversations.first()
        else:
            ai_conversation_object = get_new_conversation(request.user, conversation_type)
        return render(request, 'components/llm/ai_conversation.html', {'ai_conversation': ai_conversation_object, 'target': target, 'args': args})
        

    # Check if the conversation id is provided
    if new_conversation:
        # Check if there are any existing conversations with no messages
        ai_conversation_object = get_new_conversation(request.user, conversation_type)
        return render(request, 'components/llm/ai_conversation.html', {'ai_conversation': ai_conversation_object, 'target': target, 'args': args})

    # In cases of no conversation id
    elif not conversation_id:
        ai_conversations = AIConversation.objects.filter(
        user = request.user,
        conversation_type=conversation_type
        ).order_by('-datetime_created')

        context = {
            'ai_conversations': ai_conversations,
            'target': target
        }
        return render(request, 'components/llm/ai_conversations.html', context) 

    else:
        if delete:
            try:
                id = uuid.UUID(conversation_id)
                ai_conversation_object = AIConversation.objects.get(id=id)
                ai_conversation_object.delete()
            except:
                return HttpResponse('Invalid conversation id')

            conversations = filter_ai_conversations(request.user, conversation_type)

            return render(request, 'components/llm/ai_conversations.html', {'ai_conversations': conversations, 'target': target})        

        if rename:
            try:
                id = uuid.UUID(conversation_id)
                ai_conversation_object = AIConversation.objects.get(id=id)
                ai_conversation_object.title = rename
                ai_conversation_object.save()
            except:
                return HttpResponse('Invalid conversation id')

            conversations = filter_ai_conversations(request.user, conversation_type)

            return render(request, 'components/llm/ai_conversations.html', {'ai_conversations': conversations, 'target': target})


        # Parse uuid
        try:
            id = uuid.UUID(conversation_id)
            ai_conversation_object = AIConversation.objects.get(id=id)
        except:
            return HttpResponse('Invalid conversation id')
        
        # Get the AIConversation object
        return render(request, 'components/llm/ai_conversation.html', {'ai_conversation': ai_conversation_object, 'target': target})


def filter_ai_conversations(user, conversation_type:str):
    '''Filter AI conversations by user and conversation type.

    Args:
        user (User): The user object
        conversation_type (str): The conversation type

    Returns:
        QuerySet: The query set of AI conversations
    '''
    return AIConversation.objects.filter(
        user = user,
        conversation_type=conversation_type
    ).order_by('-datetime_created')


def get_new_conversation(user, conversation_type:str):
    '''Get a new conversation object.

    Args:
        user (User): The user object
        conversation_type (str): The conversation type

    Returns:
        AIConversation: The AI conversation object
    '''
    ai_conversations = AIConversation.objects.filter(
            user = user,
            conversation_type=conversation_type,
            conversation_history__isnull=True
        )

    if ai_conversations.exists():
        ai_conversation_object = ai_conversations.first()
    else:
        ai_conversation_object = AIConversation.objects.create(
            user = user,
            conversation_type = conversation_type
        )

    return ai_conversation_object


from django import forms

class AceEditorWidget(forms.Textarea):
    template_name = 'widgets/ace_editor_widget.html'

    def __init__(self, attrs=None, language='python'):
        attrs = attrs or {}
        attrs.setdefault('hidden', 'true')  # Hide the textarea
        super().__init__(attrs)
        self.language = language

    class Media:
        js = ('https://cdn.jsdelivr.net/npm/ace-builds@1.4.12/src-min-noconflict/ace.js',)

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget'].update({
            'language': self.language,
            'editor_id': f"editor_{attrs.get('id', name)}",
        })
        return context



from django import template
from django.db.models.manager import Manager
from django.db.models import Model
from bloomerp.utils.models import model_name_plural_underline, get_model_dashboard_view_url, get_list_view_url, get_initials, get_detail_view_url
from django.urls import reverse 
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import Link, Widget
from django.utils.safestring import mark_safe
import uuid
from bloomerp.models import Bookmark, User, ApplicationField, UserListViewPreference
from django.db.models.functions import Cast
from django.db.models import DateTimeField, F
from django.db.models import QuerySet
from bloomerp.utils.encryption import BloomerpEncryptionSuite
from django.conf import settings
from django.core.signing import dumps, loads

register = template.Library()

@register.filter(name='get_dict_value')
def get_dict_value(dictionary:dict, key:str):
    '''
    Returns the value of a key in a dictionary.

    Example usage:
    {{ dictionary|get_dict_value:key }}
    '''

    return dictionary.get(key)

@register.filter
def model_name(obj:Model):
    '''
    Returns the model name of an object.

    Example usage:
    {{ object|model_name }}
    
    '''
    return obj._meta.model_name

@register.filter
def model_name_plural(obj:Model):
    '''
    Returns the model verbose name of an object.

    Example usage:
    {{ object|model_name_plural }}
    
    '''
    return obj._meta.verbose_name_plural

@register.filter
def model_dashboard_url(content_type:ContentType):
    '''
    Returns the model app dashboard URL of an object.

    Example usage:
    {{ object|model_app_dashboard_url }}
    
    '''
    return reverse(get_model_dashboard_view_url(content_type.model_class()))

@register.filter
def model_name_plural_from_content_type(content_type:ContentType):
    '''
    Returns the model verbose name of an object.

    Example usage:
    {{ object|model_name_plural }}
    
    '''
    return content_type.model_class()._meta.verbose_name_plural

@register.filter
def length(obj) -> int:
    '''
    Returns the length of an object.

    Example usage:
    {{ object|length }}
    
    '''
    return len(obj)

@register.filter
def percentage(value, arg):
    try:
        value = int(value) / int(arg)
        return value*100
    except (ValueError, ZeroDivisionError):
        return 
    
@register.filter
def get_link_by_id(id:int):
    '''
    Returns the link object with the given id.

    Example usage:
    {{ id|get_link }}

    or 

    {% with id|get_link as link %}
    {% if link %}
    {{ link.name }}
    {% endif %}
    {% endwith %}

    '''
    try:
        return Link.objects.get(pk=id)
    except:
        return 
    
@register.filter
def get_widget_by_id(id:int):
    '''
    Returns the widget object with the given id.

    Example usage:
    {{ id|get_widget }}

    or 

    {% with id|get_widget as widget %}
    {% if widget %}
    {{ widget.name }}
    {% endif %}
    {% endwith %}

    '''
    try:
        return Widget.objects.get(pk=id)
    except:
        return 
    

@register.inclusion_tag('snippets/workspace_item.html')
def workspace_item(item:dict):
    '''
    Returns a workspace item.

    Example usage:
    {% workspace_item item %}
    '''
    # generate random id for each item
    item['id'] = uuid.uuid4()

    return {'item': item}


@register.simple_tag(name='render_link')
def render_link(link_id:int):
    '''
    Returns a link object.

    Example usage:
    {% render_link link_id %}
    '''
    try:
        link = Link.objects.get(pk=link_id)
        if link.is_external_url():
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" href="https://{link.url}" target="_blank">{link.name}</a>')
        elif link.is_absolute_url:
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" hx-get="{link.url}" hx-target="#main-content" hx-push-url="true">{link.name}</a>')
        elif not link.requires_args():
            return mark_safe(f'<a link-id="{link.pk}" class="pointer text-primary link-item" hx-get="{reverse(link.url)}" hx-target="#main-content" hx-push-url="true">{link.name}</a>')
        else:
            return mark_safe("<p>Link requires arguments</p>")
    except:
        return mark_safe("<p>Link not found</p>")

@register.inclusion_tag('components/bookmark.html')
def render_bookmark(object:Model, user:User, size:int, target:str):
    '''
    Returns a bookmark object.

    Example usage:
    {% render_bookmark object user size target %}
    '''
    # Get the content_type_id and object_id from the request
    content_type_id = ContentType.objects.get_for_model(object).pk
    
    # Check if the bookmark allready exists
    bookmarked = Bookmark.objects.filter(user=user, content_type_id=content_type_id, object_id=object.pk).exists()

    return {
        'bookmarked': bookmarked,
        'content_type_id': content_type_id,
        'object_id': object.pk,
        'target' : target,
        'size': size
    }

from bloomerp.models import File
@register.simple_tag(name='field_value')
def field_value(object:Model, application_field:ApplicationField, user:User, datatable_item:bool=False):
    '''
    Returns the formatted html value of a field in an object.
    Marks it as safe.

    Example usage:
    {% field_value object application_field user %}
    '''
    DEFAULT_NONE_VALUE = ''
    FILTER_VALUE = ''
    FILTERABLE = False

    try:
        # ------------------------------
        # GETTING THE VALUE OF THE FIELD
        # ------------------------------
        if application_field.field_type != 'OneToManyField':
            value = getattr(object, application_field.field)
            
            if value is None:
                value = DEFAULT_NONE_VALUE

        # ------------------------------
        # FORMATTING BASED ON FIELD TYPE
        # ------------------------------
        if application_field.field_type == 'ForeignKey':
            # Get the value of the field
            try:
                FILTER_VALUE = f'{application_field.field}={value.pk}' 
                FILTERABLE = True

                abosulte_url = value.get_absolute_url()
                value = mark_safe(f'<a href="{abosulte_url}">{value}</a>')
            except AttributeError:
                pass
                        
        elif application_field.field_type == 'DateField':
            # Get the date preferences of the user
            # Can be
            # ("d-m-Y", "Day-Month-Year (15-08-2000)"),
            # ("m-d-Y", "Month-Day-Year (08-15-2000)"),
            # ("Y-m-d", "Year-Month-Day (2000-08-15)"),
            preference = user.date_view_preference

            FILTER_VALUE = f'{application_field.field}={value.strftime("%Y-%m-%d")}'
            FILTERABLE = True

            # Format the date
            if preference == "d-m-Y":
                value = value.strftime("%d-%m-%Y")
            elif preference == "m-d-Y":
                value = value.strftime("%m-%d-%Y")
            elif preference == "Y-m-d":
                value = value.strftime("%Y-%m-%d")
            else:
                value = value.strftime("%d-%m-%Y")

        elif application_field.field_type == 'DateTimeField':
            # Get the datetime preferences of the user
            # Can be 
            # ("d-m-Y H:i", "Day-Month-Year Hour:Minute (15-08-2000 12:30)"),
            # ("m-d-Y H:i", "Month-Day-Year Hour:Minute (08-15-2000 12:30)"),
            # ("Y-m-d H:i", "Year-Month-Day Hour:Minute (2000-08-15 12:30)"),
            preference = user.date_view_preference

            FILTER_VALUE = f'{application_field.field}={value.strftime("%Y-%m-%d %H:%M")}'
            FILTERABLE = True

            # Format the date
            if preference == "d-m-Y H:i":
                value = value.strftime("%d-%m-%Y %H:%M")
            elif preference == "m-d-Y H:i":
                value = value.strftime("%m-%d-%Y %H:%M")
            elif preference == "Y-m-d H:i":
                value = value.strftime("%Y-%m-%d %H:%M")
            else:
                value = value.strftime("%d-%m-%Y %H:%M")

        elif application_field.field_type == 'BloomerpFileField':
            # Get the value of the field
            file:File = getattr(object, application_field.field)
            if file:
                value = mark_safe(f'<a href="{file.file.url}" target="_blank">{file.name}</a>')
            else:
                value = DEFAULT_NONE_VALUE

        elif application_field.field_type == 'ManyToManyField':
            # Get the value of the field
            qs = value.all()

            if not qs:
                value = DEFAULT_NONE_VALUE
            else:
                resp = DEFAULT_NONE_VALUE
                for item in qs[:2]:
                    resp += item.__str__() + ', '
                value = resp + '...'
        
        elif application_field.field_type == 'OneToManyField':
            # Get the value of the field
            try:
                value = getattr(object, f'{application_field.field}_set')
            except:
                value = getattr(object, application_field.field)

            qs = value.all()

            if not qs:
                value = DEFAULT_NONE_VALUE
            else:
                resp = ''
                for item in qs[:2]:
                    resp += item.__str__() + ', '
                value = resp + '...'    
        
        elif application_field.field_type == 'StatusField':
            # Get the value of the field
            filter_value : str = getattr(object, application_field.field)
            FILTER_VALUE = f'{application_field.field}={filter_value}'
            FILTERABLE = True

            # Get color
            color_dict : dict = application_field.meta.get('colors')

            if color_dict:
                color = color_dict.get(filter_value, 'gray')
            else:
                color = 'gray'

            value = mark_safe(f'<span class="badge badge-pill" style="background-color: {color}">{value}</span>')
            
            


        OTHER_FIELDS = ['AutoField', 'BigAutoField', 'BooleanField', 'CharField', 'TextField', 'IntegerField', 'DecimalField']
        if application_field.field_type in OTHER_FIELDS:
            FILTER_VALUE = f'{application_field.field}={value}'
            FILTERABLE = True

        
        # ------------------------------
        # RETURNING THE VALUE
        # ------------------------------
        if datatable_item:
            if FILTERABLE:
                return mark_safe(f'<td context-menu-filter-value="{FILTER_VALUE}" allow-context-menu={True}>{value}</td>')
            else:
                return mark_safe(f'<td allow-context-menu={True}>{value}</td>')
        else:
            return value

    except Exception as e:
        if datatable_item:
            return mark_safe(f'<td allow-context-menu={False}>{DEFAULT_NONE_VALUE}</td>')
        else:
            return DEFAULT_NONE_VALUE


@register.inclusion_tag('snippets/breadcrumb.html')
def breadcrumb(title:str=None, model:Model = None, object:Model=None):
    '''
    Returns a breadcrumb navigation.

    Example usage:
    {% breadcrumb title model object %}
    '''
    # Init context
    context = {"title": title}

    # Check if the model is set
    if model:
        list_view_url = get_list_view_url(model)
        model_dashboard_view_url = get_model_dashboard_view_url(model)
        model_name_plural = model._meta.verbose_name_plural.title()
        context['list_view_url'] = list_view_url
        context['model_name_plural'] = model_name_plural
        context['model_dashboard_url'] = model_dashboard_view_url
    if object:
        context['object'] = object
    return context



@register.inclusion_tag('snippets/avatar.html')
def avatar(object:Model, avatar_attribute:str='avatar', size:int=30, class_name=''):
    '''
    Returns an avatar object.

    Args:
        object (Model): The object that has the avatar attribute.
        avatar_attribute (str): The attribute name of the avatar. Default is 'avatar'.
        size (int): The size of the avatar. Default is 50.
        class_name (str): The class name of the avatar. Default is ''.

    Example usage:
    {% avatar object avatar_attribute size class_name %}

    '''
    try:
        avatar = getattr(object, avatar_attribute)

        if not hasattr(avatar, 'url'):
            # Get the first letter of the object's string representation
            initials = get_initials(object)
        else:
            initials = None
    except:
        initials = get_initials(object)
        avatar = None

    return {
        'avatar': avatar,
        'size': size,
        'class_name': class_name,
        'initials': initials
    }


import uuid
@register.inclusion_tag('snippets/datatable_and_filter.html')
def datatable(
    content_type_id:int,
    user:User,
    include_actions:bool=True,
    initial_query:str='',
    request=None,
    datatable_id=None,
    bypass_view_permission=False
):
    '''
    Returns a data table for a model.

    Example usage:
    {% datatable content_type_id user include_actions initial_query request datatable_id bypass_view_permission %}
    '''
    # Get the model from the content_type_id
    content_type = ContentType.objects.get(pk=content_type_id)
    model = content_type.model_class()

    # Get the application fields
    application_fields = ApplicationField.objects.filter(content_type=content_type)

    # Create random id for the datatable target
    if not datatable_id:
        datatable_id = 'datatable-' + str(uuid.uuid4())
    else:
        datatable_id = 'datatable-' + str(datatable_id)

    if bypass_view_permission:
        if not initial_query:
            raise ValueError('Initial query must be set if bypass_view_permission is True')

        bypass_view_permission_value = dumps({
            'initial_query' : initial_query,
            'content_type_id': content_type_id,
            'user_id' : user.pk
        })

    # Add the bypass_view_permission_value to the initial_query
    if initial_query and bypass_view_permission:
        initial_query += f'&data_table_bypass_view_permission={bypass_view_permission_value}'
    elif bypass_view_permission and not initial_query:
        initial_query = f'data_table_bypass_view_permission={bypass_view_permission_value}'


    return {
        'model': model,
        'application_fields': application_fields,
        'content_type_id': content_type_id,
        'datatable_id': datatable_id,
        'include_actions': include_actions,
        'initial_query': initial_query,
        'request': request,
    }


@register.simple_tag(takes_context=True)
def generate_uuid(context):
    '''
    Returns a unique id.
    '''
    return str(uuid.uuid4())


@register.filter
def detail_view_url(object:Model):
    '''
    Returns the absolute url of an object.

    Example usage:
    {{ object|detail_view_url }}

    '''
    try:
        return object.get_absolute_url()
    except:
        model = object._meta.model
        return reverse(get_detail_view_url(model), kwargs={'pk': object.pk})


@register.inclusion_tag("snippets/calendar.html")
def calendar(queryset: QuerySet, start_date_field: str, end_date_field: str = None, id=None):
    if not id:
        id = str(uuid.uuid4())

    parse = 'Y-m-d'
    queryset = queryset.annotate(date_start=Cast(F(start_date_field), output_field=DateTimeField()))
    if end_date_field:
        queryset = queryset.annotate(date_end=Cast(F(end_date_field), output_field=DateTimeField()))


    return {
        "queryset": queryset,
        "start_date_field": start_date_field,
        "end_date_field": end_date_field,
        "id": id,
        "parse":parse
    }



from django.views.generic import DetailView
from django.contrib.auth.mixins import PermissionRequiredMixin
from bloomerp.models import DocumentTemplate, File, DocumentTemplateFreeVariable, DocumentTemplateStyling, DocumentTemplateHeader
from bloomerp.forms.document_templates import FreeVariableForm, GenerateDocumentForm
from django.shortcuts import redirect
from django.contrib import messages
from bloomerp.utils.document_templates import DocumentController
from django.contrib.contenttypes.models import ContentType
from bloomerp.utils.router import BloomerpRouter
from bloomerp.views.core import BloomerpBaseDetailView
from django.views.generic.edit import UpdateView
from bloomerp.utils.models import get_detail_base_view_url


# ---------------------------------
# Bloomerp Detail Document Template List View
# ---------------------------------
router = BloomerpRouter()

EXCLUDE_MODELS = [
    DocumentTemplate,
    DocumentTemplateFreeVariable,
    DocumentTemplateStyling,
    DocumentTemplateHeader
]


@router.bloomerp_route(
    path="document-templates", 
    name="Document Templates",
    url_name="document_templates_list",
    description="List of document templates for the {model} model",
    route_type="detail",
    exclude_models=EXCLUDE_MODELS
    )
class BloomerpDetailDocumentTemplateListView(PermissionRequiredMixin, BloomerpBaseDetailView):
    settings = None
    template_name = "document_template_views/bloomerp_document_template_list_view.html"
    document_template_detail_url = "document_template_detail_editor" # URL to view document template detail
    document_template_generate_url = "" # URL to generate document template for a specific instance

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get related document templates
        content_type = ContentType.objects.get_for_model(self.model)

        context["document_templates"] = DocumentTemplate.objects.filter(
            model_variable=content_type
        )
        context["document_template_detail_url"] = self.document_template_detail_url
        context["document_template_generate_url"] = get_detail_base_view_url(self.model) + "_document_template_generate"
        return context

# ---------------------------------
# Bloomerp Detail Document Template Generate View
# ---------------------------------

@router.bloomerp_route(
    path="document-templates/<int:template_id>/", 
    name="Document Template generate for {model}",
    route_type="detail",
    url_name="document_template_generate",
    description="Document Template for {model}",
    exclude_models=EXCLUDE_MODELS
    )
class BloomerpDetailDocumentTemplateGenerateView(PermissionRequiredMixin, BloomerpBaseDetailView):
    template_name = "document_template_views/bloomerp_detail_document_generator_view.html"
    model = None

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

    def post(self, request, *args, **kwargs):
        # Retrieve the instance based on the provided ID
        template_id = self.kwargs["template_id"]
        # Use the retrieved ID as desired
        template = DocumentTemplate.objects.get(pk=template_id)
        
        try:
            if "instance_select" in request.POST:
                id = request.POST.get("instance_select")
                instance = template.model_variable.get_object_for_this_type(pk=id)
            else:
                instance = self.get_object()

            free_variable_form = FreeVariableForm(template, request.POST)

            if free_variable_form.is_valid():
                data = free_variable_form.cleaned_data
                controller = DocumentController(
                    document_template=template,
                    user=self.request.user
                )
                controller.create_document(template, instance, data)

            # Redirect to a success page or any other desired view
        except Exception as e:
            # Handle the exception
            messages.error(request, f"Document error: {e}")
            # You might want to log the error or provide appropriate feedback to the user

        return redirect(request.path)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        instance = self.get_object()

        # Retrieve document, important that the kwarg = id
        template_id = self.kwargs["template_id"]
        template = DocumentTemplate.objects.get(pk=template_id)

        context["document_template"] = template
        context["free_variable_form"] = FreeVariableForm(document_template=template)

        # Retrieve file
        file_queryset = File.objects.filter(
            meta__document_template=template_id,
            object_id=instance.pk,
            content_type_id=ContentType.objects.get_for_model(self.model),
        ).order_by("-datetime_created")

        context["file_list"] = file_queryset

        return context




@router.bloomerp_route(
    models = DocumentTemplate, 
    path="editor",
    route_type="detail", 
    name="Editor",
    url_name="editor",
    description="Document Template Editor"
    )
class BloomerpDocumentTemplateEditorView(PermissionRequiredMixin, BloomerpBaseDetailView, UpdateView):
    model = DocumentTemplate
    template_name = "document_template_views/bloomerp_document_template_editor_view.html"
    fields = ["template"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["args"] = f"template_id={self.object.id}" 
        return context

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

@router.bloomerp_route(
    models = DocumentTemplate, 
    path="generate-document",
    route_type="detail", 
    name="Generate document",
    url_name="generate_document",
    description="Generate document for the document template"
    )
class BloomerpDocumentTemplateGenerateView(PermissionRequiredMixin, BloomerpBaseDetailView):
    model = DocumentTemplate
    template_name = "document_template_views/bloomerp_document_template_generate_view.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Check if document template has 
        
        context['form'] = GenerateDocumentForm(self.object)

        return context
    
    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]

@router.bloomerp_route(
    models = DocumentTemplate, 
    path="generated-documents",
    route_type="detail", 
    name="Generated documents",
    url_name="generated_documents",
    description="List of generated documents for the document template"
    )
class BloomerpDocumentTemplateGeneratedDocumentsView(PermissionRequiredMixin, BloomerpBaseDetailView):
    model = DocumentTemplate
    template_name = "document_template_views/bloomerp_document_template_generated_documents_view.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Get all files associated with the document template
        files = File.objects.filter(meta__document_template=self.get_object().pk)
        context['files'] = files

        return context

    def get_permission_required(self):
        return [f"{self.model._meta.app_label}.view_{self.model._meta.model_name}"]




# Generated by Django 5.1.1 on 2025-01-08 01:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0065_alter_todo_priority_alter_todo_required_by"),
    ]

    operations = [
        migrations.RenameField(
            model_name="todo",
            old_name="requested_for",
            new_name="assigned_to",
        ),
    ]



from django.http import HttpResponse
from django.template.loader import render_to_string
from django.core.exceptions import PermissionDenied

class HTMXPermissionDeniedMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = None
        try:
            response = self.get_response(request)
        except Exception as e:
            response = self.process_exception(request, e)
            if response is None:
                raise
        return response

    def process_exception(self, request, exception):
        if isinstance(exception, PermissionDenied):
            if request.headers.get('HX-Request'):
                response_html = render_to_string('snippets/403.html')
                return HttpResponse(response_html, status=200)
            else:
                response_html = render_to_string('403.html')
                return HttpResponse(response_html, status=403)
                
        return None


# Generated by Django 5.1.1 on 2025-01-11 02:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0069_alter_aiconversation_conversation_history"),
    ]

    operations = [
        migrations.AlterField(
            model_name="comment",
            name="object_id",
            field=models.CharField(max_length=36),
        ),
        migrations.AlterField(
            model_name="file",
            name="object_id",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
        migrations.AlterField(
            model_name="todo",
            name="object_id",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-03 21:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0027_filefolder_created_by_filefolder_datetime_created_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="filefolder",
            name="files",
            field=models.ManyToManyField(
                blank=True, null=True, related_name="folders", to="bloomerp.file"
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-10 14:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0043_applicationfield_db_column"),
    ]

    operations = [
        migrations.CreateModel(
            name="Workspace",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The name of the workspace", max_length=255
                    ),
                ),
                (
                    "content",
                    models.JSONField(
                        default={
                            "content": [
                                {
                                    "data": {"text": "Welcome to your workspace"},
                                    "size": 12,
                                    "type": "header",
                                },
                                {
                                    "data": {
                                        "text": "This is your workspace. You can add widgets, links, and other content here."
                                    },
                                    "size": 12,
                                    "type": "text",
                                },
                                {
                                    "data": {"text": "Example of widget"},
                                    "size": 12,
                                    "type": "header",
                                },
                                {
                                    "data": {"widget_id": 5},
                                    "size": 12,
                                    "type": "widget",
                                },
                                {
                                    "data": {"text": "Example of link"},
                                    "size": 12,
                                    "type": "header",
                                },
                                {"data": {"link_id": 1}, "size": 12, "type": "link"},
                                {
                                    "data": {"text": "Example of link"},
                                    "size": 12,
                                    "type": "header",
                                },
                                {"data": {"id": 2}, "size": 12, "type": "link"},
                                {
                                    "data": {"text": "Example of link list"},
                                    "size": 12,
                                    "type": "header",
                                },
                                {
                                    "data": {"links": [1, 2]},
                                    "size": 12,
                                    "type": "link_list",
                                },
                            ]
                        },
                        help_text="The content of the workspace",
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_workspace",
                "managed": True,
            },
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.models import ApplicationField
from bloomerp.forms.core import BloomerpModelForm
from django.forms.models import modelform_factory
import uuid
from django.contrib.auth.decorators import login_required

@login_required
@route('list_view_filter_inputs')
def list_view_filter_inputs(request:HttpRequest) -> HttpResponse:
    '''
    Returns the list view filter inputs:

    Variables:
        - application_field_id: The application field id
        - step :  Can be either 'action' or 'input'
        - action : The filter action to be performed
    '''
    # Initialize variables
    form = None
    EXCLUDED_FIELD_TYPES = ['Property', 'OneToManyField'] # Field types to exclude in the filter mechanism


    # Get the get parameters
    content_type_id = request.GET.get('content_type_id', None)
    application_field = request.GET.get('application_field_id', None)
    step = request.GET.get('step', None)

    # Some validation
    if step not in ['action', 'input','row']:
        return HttpResponse("Step is required and must be either 'action' or 'input' or 'row'", status=400)
    
    if step in ['action', 'input']:
        if not application_field:
            return HttpResponse("Application field is required", status=400)
        else:
            # Get the application field
            application_field = ApplicationField.objects.get(id=application_field)
            model = application_field.content_type.model_class()

    # Action step: renders the specific actions that can be performed on the field
    if step == 'action':
        return render(request, 'components/list_view_filter_action.html', {'application_field': application_field,'id':uuid.uuid4()})
    
    # Input step: renders the input field for the filter
    if step == 'input':
        context = {
            'application_field': application_field,
        }

        # Get the action parameter
        action = request.GET.get('action', None)
        if not action:
            return HttpResponse("Action is required", status=400)
        else:
            context['action'] = action

        # If the field is a foreign key or many to many field, get the related fields
        if application_field.field_type in ['ForeignKey', 'ManyToManyField']:
            if action == 'advanced':
                related_application_fields = ApplicationField.objects.filter(content_type=application_field.related_model).exclude(field_type__in=EXCLUDED_FIELD_TYPES)
                context['related_application_fields'] = related_application_fields

            else:
                # Create a model form that only includes the related field
                form = modelform_factory(model=model, form=BloomerpModelForm, fields=[application_field.field])(model=model, user=request.user, apply_helper=False, hide_default_fields=False)
                
                
                context['form'] = form

        return render(request, 'components/list_view_filter_input.html', context)
    
    # Row step: renders the row filter
    elif step == 'row':
        # Requires content_type_id
        if not content_type_id:
            return HttpResponse("Content type id is required", status=400)
        
        application_fields = ApplicationField.objects.filter(content_type_id=content_type_id).exclude(field_type__in=EXCLUDED_FIELD_TYPES)

        return render(request, 'components/list_view_filter_row.html', {'filter_application_fields': application_fields, 'id':uuid.uuid4()})





# Generated by Django 5.1.1 on 2024-10-30 18:12

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0038_delete_userdetailviewtabs"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserDetailViewTab",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "link",
                    models.ForeignKey(
                        help_text="The link to be displayed in the detail view tab",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bloomerp.link",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_detail_view_tab",
                "managed": True,
                "unique_together": {("user", "content_type", "link")},
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-09-22 23:24

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0009_rename_created_at_accommodation_datetime_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="customer",
            options={"verbose_name": "customer", "verbose_name_plural": "customers"},
        ),
    ]



from langchain_core.tools import tool, StructuredTool, BaseTool
from bloomerp.utils.sql import SqlQueryExecutor
from bloomerp.models import ApplicationField, User, Link
    

class BaseBloomerpTool(BaseTool):
    requires_user : bool = False # Tool that requires user id
    user : User = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = kwargs.get('user', None)

    


class GetDatabaseTablesTool(BaseBloomerpTool):
    description : str = '''
                    Function that returns a dictionary of tables and their columns in a database
                    Can be used to get to know the structure of the database for further querying
                    Function should only be called once during every conversation.
                    Tables that are not accessible to the user will not be returned.
                    '''
    name : str = "get_accessible_database_tables"


    def _run(self) -> list[tuple[str, list[str]]]:
        tables = ApplicationField.get_db_tables_and_columns(self.user)
        if not tables:
            return "User does not have access to any tables"
        else:
            return tables
    

class PerformSqlQueryTool(BaseBloomerpTool):
    description : str = '''Function that performs an SQL query and returns the result.
                    Can only be used to perform SELECT queries.'''
    name : str = "perform_sql_query"

    def _run(self, query: str) -> list[list[str]]:
        sql_executor = SqlQueryExecutor()
        try:
            result = sql_executor.execute_raw(query)
            return result
        except Exception as e:
            return [[str(e)]]
    
class GetApplicationRoutesTool(BaseBloomerpTool):
    description : str = '''Function that returns the list of application routes that are accessible to the user.
    Args:
        key_word : list[str] - A list of keywords to filter the routes
    '''
    name : str = "get_application_routes"

    def _run(self, key_words: list[str]) -> list[str]:
        links = []

        qs = Link.objects.none()
        for key_word in key_words:
            qs = qs | Link.objects.filter(name__icontains=key_word)

        qs = qs.distinct()

        for link in qs:
            if not link.requires_args():
                links.append(link.to_absolute_url())

        return links
            


        


BLOOMAI_TOOLS = [PerformSqlQueryTool, GetDatabaseTablesTool, GetApplicationRoutesTool]  # List of tools that can be used in the bloomai


# Generated by Django 5.1.1 on 2024-11-20 21:34

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0016_alter_employee_manager"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="accommodation",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="employee",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="employeeonaccommodation",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="equipment",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="function",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "permissions": (
                    ("bulk_delete", "Can delete multiple objects from at once"),
                    ("bulk_change", "Can update multiple objects from at once"),
                )
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-11 11:28

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0045_remove_userdashboardview_created_by_and_more"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="workspace",
            name="content_type",
            field=models.ForeignKey(
                blank=True,
                help_text="The content type related to the workspace",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AddField(
            model_name="workspace",
            name="user",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.1.1 on 2024-11-11 11:22

import bloomerp.models.auth
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0044_workspace"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="userdashboardview",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="userdashboardview",
            name="updated_by",
        ),
        migrations.RemoveField(
            model_name="userdashboardview",
            name="user",
        ),
        migrations.RemoveField(
            model_name="userdashboardview",
            name="widgets",
        ),
        migrations.AlterUniqueTogether(
            name="userkpidashboardviewpreference",
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name="userkpidashboardviewpreference",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="userkpidashboardviewpreference",
            name="updated_by",
        ),
        migrations.RemoveField(
            model_name="userkpidashboardviewpreference",
            name="user",
        ),
        migrations.RemoveField(
            model_name="userkpidashboardviewpreference",
            name="widget",
        ),
        migrations.AlterField(
            model_name="workspace",
            name="content",
            field=models.JSONField(
                default=bloomerp.models.auth.get_default_workspace,
                help_text="The content of the workspace",
            ),
        ),
        migrations.DeleteModel(
            name="UserAppDashboard",
        ),
        migrations.DeleteModel(
            name="UserDashboardView",
        ),
        migrations.DeleteModel(
            name="UserKpiDashboardViewPreference",
        ),
    ]



from django import forms
from django.forms.models import modelform_factory
from django.db.models import Model
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import ApplicationField
import random

class ForeignKeyWidget(forms.Widget):
    template_name = 'widgets/foreign_field_widget.html'

    def __init__(self, model: Model, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
    
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        # Get first 5 objects from the model
        context['objects'] = self.model.objects.all()[:5]
        context['content_type_id'] = ContentType.objects.get_for_model(self.model).id


        # Create form for the model to be used in the widget
        Form = modelform_factory(self.model, fields='__all__')
        form_prefix = context['widget']['name']
        context['form_prefix'] = form_prefix
        context['form'] = Form(prefix=form_prefix)

        # Add random modal id to the context
        context['advanced_search_modal_id'] = f'modal-{random.randint(0, 100000)}'
        context['create_modal_id'] = f'modal-{random.randint(0, 100000)}'

        # Get the object that is currently selected
        if context['widget']['value']:
            context['selected_object'] = self.model.objects.get(pk=context['widget']['value'])

        # Check if an invalid entry was made
        if attrs.get('aria-invalid', 'false') == 'true':
            context['invalid'] = True

        # Add application fields to the context
        context['application_fields'] = ApplicationField.objects.filter(content_type=ContentType.objects.get_for_model(self.model)).exclude(field_type='Property')

        context['widget_type'] = 'fk'
        return context



# Generated by Django 5.1.1 on 2024-10-02 22:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0023_userdashboardview_layout"),
    ]

    operations = [
        migrations.CreateModel(
            name="Folder",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "files",
                    models.ManyToManyField(related_name="folders", to="bloomerp.file"),
                ),
            ],
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.shortcuts import get_object_or_404
from bloomerp.models import Widget
import plotly.express as px
from django.contrib.auth.decorators import login_required
from django.core.cache import cache

@login_required
@route('workspace_widget')
def workspace_widget(request:HttpRequest) -> HttpResponse:
    '''
    Component to render a widget in the workspace.

    GET Parameters:
    - widget_id: The ID of the widget to render
    - download: If set to xlsx or csv, the component will return a download response
    '''
    # Permission check
    if not request.user.has_perm('bloomerp.view_workspace'):
        return HttpResponse(
            'User does not have permission to view widgets'
        )


    # Get the KPI ID
    widget_id = request.GET.get('widget_id')
    if not widget_id:
        return HttpResponse('Widget ID is required', status=400)
    
    # Check if the a refresh is requested
    # Refresh yet to be implemented 
    refresh = request.GET.get('refresh', False)
    if refresh:
        cache.delete(f'widget_{widget_id}')


    # Check cache
    cache_key = f'widget_{widget_id}'
    cached_response = cache.get(cache_key)
    if cached_response:
        return cached_response

    # Get download
    download = request.GET.get('download', False)

    # Get the KPI object
    widget = get_object_or_404(Widget, pk=widget_id)

    # Get the data
    x = widget.options.get('x')
    y = widget.options.get('y')
    group_by = widget.options.get('group_by')
    data = widget.query.executor.execute_to_df(
        query=widget.query.query,
        safe=True,
        use_cache=True
    )


    #--------------------------------
    # Download
    #--------------------------------
    if download == 'xlsx':
        response = HttpResponse(content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        response['Content-Disposition'] = f'attachment; filename="{widget.name}.xlsx"'
        data.to_excel(response, index=False)
        return response
    elif download == 'csv':
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{widget.name}.csv"'
        data.to_csv(response, index=False)
        return response

    #--------------------------------
    # Output to HTML
    #--------------------------------

    if widget.output_type == 'value':
        result = data.iloc[0, 0]
    elif widget.output_type == 'bar':
        fig = px.bar(data, x=x, y=y, color=group_by, height=300)
        result = fig.to_html(full_html=False)
    elif widget.output_type == 'line':
        fig = px.line(data, x=x, y=y, color=group_by, height=300)
        result = fig.to_html(full_html=False)
    elif widget.output_type == 'pie':
        fig = px.pie(data, names=group_by, values=x, height=300)
        result = fig.to_html(full_html=False)
    elif widget.output_type == 'scatter':
        fig = px.scatter(data, x=x, y=y, color=group_by, height=300)
        result = fig.to_html(full_html=False)
    elif widget.output_type == 'histogram':
        fig = px.histogram(data, x=x, color=group_by, height=300)
        result = fig.to_html(full_html=False)

    elif widget.output_type == 'table':
        columns = widget.options.get('columns')
        limit = widget.options.get('limit')
        data = data[columns].iloc[:limit].values.tolist()
        result = {
            'columns': columns,
            'data': data
        }

    # Add context
    context = {
        'widget': widget,
        'result': result
    }
    
    response = render(request, 'components/workspace_widget.html', context)
    
    # Set cache
    cache.set(cache_key, response, timeout=60*15)  # Cache for 15 minutes
    
    return response


# Generated by Django 5.1.1 on 2025-01-02 22:55

import bloomerp.models.fields
import bloomerp_tests.models
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp_tests", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="child",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                    "export",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="parent",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                    "export",
                )
            },
        ),
        migrations.RemoveField(
            model_name="child",
            name="parent",
        ),
        migrations.AddField(
            model_name="child",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="child",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="child",
            name="datetime_created",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="child",
            name="datetime_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="child",
            name="parent_with_related_name",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="bloomerp_tests.parent",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="child",
            name="parent_without_related_name",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                to="bloomerp_tests.parent",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="child",
            name="status",
            field=bloomerp.models.fields.StatusField(
                choices=[
                    ("active", "Active"),
                    ("inactive", "Inactive"),
                    ("pending", "Pending"),
                ],
                colored_choices=[
                    ("active", "Active", "#00ff00"),
                    ("inactive", "Inactive", "#ff0000"),
                    ("pending", "Pending", "#ffff00"),
                ],
                default="active",
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name="child",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="parent",
            name="avatar",
            field=models.ImageField(blank=True, null=True, upload_to="avatars/"),
        ),
        migrations.AddField(
            model_name="parent",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="parent",
            name="date",
            field=models.DateField(default=bloomerp_tests.models.generate_random_date),
        ),
        migrations.AddField(
            model_name="parent",
            name="datetime_created",
            field=models.DateTimeField(auto_now_add=True, default=1),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="parent",
            name="datetime_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="parent",
            name="status",
            field=bloomerp.models.fields.StatusField(
                choices=[
                    ("active", "Active"),
                    ("inactive", "Inactive"),
                    ("pending", "Pending"),
                ],
                colored_choices=[
                    ("active", "Active", "#00ff00"),
                    ("inactive", "Inactive", "#ff0000"),
                    ("pending", "Pending", "#ffff00"),
                ],
                default="active",
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name="parent",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="Many",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                (
                    "avatar",
                    models.ImageField(blank=True, null=True, upload_to="avatars/"),
                ),
                ("name", models.CharField(max_length=255)),
                ("age", models.IntegerField(default=71)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "date",
                    models.DateField(
                        default=bloomerp_tests.models.generate_random_date
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(class)s_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                    "export",
                ),
            },
        ),
        migrations.AddField(
            model_name="child",
            name="many_to_many_with_related_name",
            field=models.ManyToManyField(
                related_name="children", to="bloomerp_tests.many"
            ),
        ),
        migrations.AddField(
            model_name="child",
            name="many_to_many_without_related_name",
            field=models.ManyToManyField(to="bloomerp_tests.many"),
        ),
    ]



# Generated by Django 5.1.1 on 2025-01-08 00:09

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0063_rename_user_todo_requested_for_todo_title_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="todo",
            name="avatar",
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-29 18:06

import bloomerp.models.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0026_remove_employee_accommodations"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="status",
            field=bloomerp.models.fields.StatusField(
                choices=[
                    ("pending", "Pending"),
                    ("processing", "Processing"),
                    ("completed", "Completed"),
                    ("cancelled", "Cancelled"),
                ],
                colored_choices=[
                    ("pending", "Pending", "#ffcc00"),
                    ("processing", "Processing", "#007bff"),
                    ("completed", "Completed", "#28a745"),
                    ("cancelled", "Cancelled", "#dc3545"),
                ],
                max_length=30,
            ),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render, redirect
from django.http import HttpResponse, HttpRequest
from bloomerp.forms.core import DetailLinksSelectForm
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required

@login_required
@route('detail_view_links')
def detail_view_links(request:HttpRequest) -> HttpResponse:
    '''Add component to select detail links for a user'''

    # Get the content type id from the query parameters
    content_type_id = request.POST.get('content_type_id')
    if not content_type_id:
        return HttpResponse('content_type_id required in the query parameters', status=400)
    else:
        ct = ContentType.objects.get(id=content_type_id)
    
    form = DetailLinksSelectForm(
        data =request.POST,
        user = request.user,
        content_type=ct
        )
    
    if form.is_valid():
        form.save()

    # Redirect to the page the url came from
    url = request.META.get('HTTP_REFERER')

    return redirect(url)


import io
from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import FileResponse, HttpRequest, HttpResponse
from django.contrib.contenttypes.models import ContentType
from bloomerp.utils.model_io import BloomerpModelIO
from bloomerp.forms.core import BloomerpDownloadBulkUploadTemplateForm
from django.contrib.auth.decorators import login_required

@login_required
@route('download_bulk_upload_template')
def download_bulk_upload_template(request: HttpRequest) -> FileResponse:
    '''Save the bulk upload template for a model as a CSV or Excel file.'''
    form = BloomerpDownloadBulkUploadTemplateForm(data=request.POST)

    if not form.is_valid():
        # Print form errors explicitly
        print("Form errors:", form.errors)
        return HttpResponse('Invalid form data', status=400)

    # Get the selected fields from the form
    fields = form.get_selected_fields()
    
    file_type = form.cleaned_data.get('file_type')
    model = form.model

    # Permission check
    if not request.user.has_perm(f'{model._meta.app_label}.add_{model._meta.model_name}'):
        return HttpResponse('User does not have permission to download the bulk upload template', status=403)

    if not model:
        return HttpResponse('Model not found', status=400)

    model_io = BloomerpModelIO(model)

    if file_type == 'csv':
        template_bytes = model_io.create_model_template(file_type='csv', fields=fields)
        content_type = 'text/csv'
        extension = 'csv'
    elif file_type == 'xlsx':
        template_bytes = model_io.create_model_template(file_type='xlsx', fields=fields)
        content_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        extension = 'xlsx'
    else:
        return HttpResponse('Invalid export format', status=400)

    # Wrap the bytes in a BytesIO object
    byte_stream = io.BytesIO(template_bytes)

    # Explicitly set the Content-Disposition header
    response = HttpResponse(byte_stream, content_type=content_type)
    response['Content-Disposition'] = f'attachment; filename={model.__name__}__bulk_upload_template.{extension}'

    return response


# Generated by Django 5.1.1 on 2024-09-15 22:05

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="customer",
            options={"verbose_name": "Customer", "verbose_name_plural": "Customers"},
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "permissions": [
                    ("bulk_delete", "Can delete multiple objects at once"),
                    ("bulk_update", "Can update multiple objects at once"),
                ]
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "permissions": [
                    ("bulk_delete", "Can delete multiple objects at once"),
                    ("bulk_update", "Can update multiple objects at once"),
                ]
            },
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-06 18:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0033_userappdashboard"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userappdashboard",
            name="layout",
            field=models.JSONField(
                default={}, help_text="The layout of the app dashboard"
            ),
            preserve_default=False,
        ),
    ]



from django.db import connection
from django.core.cache import cache
import re
import pandas as pd
from typing import Any, List, Dict
import time

class SqlQueryExecutor:

    def __init__(self, cache_time: int = 60, cache_id: str = None):
        self.cache_time = cache_time
        self.cache_id = cache_id

    def execute_raw(self, query: str, safe: bool = True, use_cache: bool = False) -> List:
        """
        Execute a query and return the result as a list of rows.
        
        Parameters:
            - query: The SQL query to execute.
            - safe: If True, the query must be a SELECT statement, otherwise it will raise a ValueError.
            - use_cache: If True, the result will be fetched from cache if available.
        
        Returns:
            A list of rows from the executed query.
        """
        self._validate_safe_query(query, safe)
        
        if use_cache:
            cache_result = self._get_cache_result()
            if cache_result:
                return cache_result

        with connection.cursor() as cursor:
            cursor.execute(query)
            result = cursor.fetchall()

        if use_cache:
            self._set_cache_result(result)

        return result

    def execute_to_dict(self, query: str, safe: bool = True, use_cache: bool = False) -> Dict[str, List]:
        """
        Execute a query and return the result as a dictionary with columns and rows.
        
        Parameters:
            - query: The SQL query to execute.
            - safe: If True, the query must be a SELECT statement, otherwise it will raise a ValueError.
            - use_cache: If True, the result will be fetched from cache if available.
        
        Returns:
            A dictionary with 'columns' and 'rows' keys.
        """
        self._validate_safe_query(query, safe)
        
        if use_cache:
            cache_result = self._get_cache_result()
            if cache_result:
                return cache_result

        with connection.cursor() as cursor:
            cursor.execute(query)
            columns = [col[0] for col in cursor.description]
            rows = cursor.fetchall()
            result = {"columns": columns, "rows": rows}

        if use_cache:
            self._set_cache_result(result)

        return result

    def execute_to_first_value(self, query: str, safe: bool = True, use_cache: bool = False) -> Any:
        """
        Execute a query and return the first value of the first row.
        
        Parameters:
            - query: The SQL query to execute.
            - safe: If True, the query must be a SELECT statement, otherwise it will raise a ValueError.
            - use_cache: If True, the result will be fetched from cache if available.
        
        Returns:
            The first value of the first row from the executed query.
        """
        self._validate_safe_query(query, safe)
        
        if use_cache:
            cache_result = self._get_cache_result()
            if cache_result:
                return cache_result

        with connection.cursor() as cursor:
            cursor.execute(query)
            result = cursor.fetchone()[0]

        if use_cache:
            self._set_cache_result(result)

        return result

    def execute_to_df(self, query: str, safe: bool = True, use_cache: bool = False) -> pd.DataFrame:
        """
        Execute a query and return the result as a pandas DataFrame.
        
        Parameters:
            - query: The SQL query to execute.
            - safe: If True, the query must be a SELECT statement, otherwise it will raise a ValueError.
            - use_cache: If True, the result will be fetched from cache if available.
        
        Returns:
            A pandas DataFrame containing the query result.
        """
        self._validate_safe_query(query, safe)
        
        if use_cache:
            cache_result = self._get_cache_result()
            if cache_result is not None:
                return cache_result

        with connection.cursor() as cursor:
            cursor.execute(query)
            columns = [col[0] for col in cursor.description]
            result = pd.DataFrame(cursor.fetchall(), columns=columns)
            

        if use_cache:
            self._set_cache_result(result)

        return result

    def is_safe(self, query: str) -> bool:
        """
        Check if a query is safe to execute. Safe queries are SELECT statements.
        
        Parameters:
            - query: The SQL query to check.
        
        Returns:
            True if the query is safe, False otherwise.
        """
        return not re.search(r'\b(update|delete|insert|drop|alter|create|grant|revoke|truncate)\b', query.lower())

    def is_valid(self, query: str) -> bool:
        """
        Check if a query is valid by attempting to execute it.
        
        Parameters:
            - query: The SQL query to check.
        
        Returns:
            True if the query is valid, raises an exception otherwise.
        """
        try:
            with connection.cursor() as cursor:
                cursor.execute(query)
            return True
        except Exception as e:
            return e

    def get_query_from_cache(self, cache_id: str) -> str:
        """
        Retrieve a query result from the cache.
        
        Parameters:
            - cache_id: The cache id to retrieve the query result.
        
        Returns:
            The cached query result.
        """
        return cache.get(cache_id)

    def _validate_safe_query(self, query: str, safe: bool):
        if not self.is_safe(query) and safe:
            raise ValueError('Unsafe query')

    def _get_cache_result(self) -> Any:
        """
        Retrieve the cached result using the cache_id.
        """
        return cache.get(self.cache_id)

    def _set_cache_result(self, result: Any):
        """
        Set the result in cache with the cache_id and cache_time.
        """
        cache.set(self.cache_id, result, self.cache_time)




from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.forms.core import BloomerpModelForm
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required
from django.forms.models import modelform_factory
from django.contrib import messages
from bloomerp.models import File
from django.shortcuts import get_object_or_404

@login_required
@route('bulk_update_objects')
def bulk_update_objects(request:HttpRequest) -> HttpResponse:
    # Get the content type id and form prefix from the query parameters
    content_type_id = request.GET.get('content_type_id')
    form_prefix = request.GET.get('bulk_update_form_prefix','')
    user = request.user
    num_objects = 0


    # Initialize potential errors
    errors = []

    if request.method == 'POST':
        # Get objects to update
        object_ids = request.POST.getlist('object_ids')
        ct = get_object_or_404(ContentType, pk=content_type_id)
        model = ct.model_class()
        objects = model.objects.filter(id__in=object_ids)
        delete_objects = request.POST.get('delete_objects', False)            
        

        # Number of objects
        num_objects = len(objects)
        num_errors = 0

        # If delete_objects has a value
        if delete_objects in ['true', 'True']:
            if not user.has_perm(f'{model._meta.app_label}.bulk_delete_{model._meta.model_name}'):
                return HttpResponse('Permission denied', status=403)

            # Check whether the model is a file
            if model == File:
                for obj in objects:
                    obj.delete()
            else:
                objects.delete()
            messages.success(request, f"Successfully deleted {num_objects} objects")
        else:
            # If the user does not have permission to update the model
            if not user.has_perm(f'{model._meta.app_label}.bulk_change_{model._meta.model_name}'):
                return HttpResponse('Permission denied', status=403)

            # Get all of the fields that are being updated from request.POST using the form_prefi
            fields = []
            if not form_prefix:
                return HttpResponse('form_prefix required in the query parameters', status=400)
            x = len(form_prefix) + 1 # The length of the form_prefix plus the underscore Ex. 'form_prefix_'
            for key in request.POST:
                if key.startswith(form_prefix):
                    fields.append(key[x:])
            
            # Create a formset with the fields
            Form = modelform_factory(model, fields=fields, form=BloomerpModelForm)

            # Loop through the objects and update them
            
            for obj in objects:
                form = Form(data=request.POST, files=request.FILES, prefix=form_prefix, instance=obj, model=model, user=user)
                if form.is_valid():
                    form.save()
                else:
                    errors.append(
                        {
                            "object" : obj,
                            "errors" : form.errors
                        }
                    )
            # Get number of errors
            num_errors = len(errors)
            if num_errors == num_objects:
                messages.error(request, f"Failed to update {num_errors} objects")
            elif num_errors > 0:
                messages.warning(request, f"Bulk update paritally completed. Failed to update {num_errors} objects")
            else:
                messages.success(request, f"Successfully updated {num_objects} objects")

    context = {
        "errors": errors,
        "num_objects": num_objects,
        "num_errors": num_errors
    }
    return render(request, 'components/bulk_update_objects.html', context)


from bloomerp.forms.core import BloomerpModelForm, DetailLinksSelectForm
from django.shortcuts import redirect
from django.http import HttpResponse
from django.forms.models import modelform_factory
from bloomerp.models import ApplicationField, UserDetailViewTab, Link
from django.views.generic import DetailView, UpdateView
from bloomerp.utils.models import (
    get_create_view_url,
    get_update_view_url,
    get_list_view_url,
    get_model_dashboard_view_url,
    get_detail_view_url,
    get_bulk_upload_view_url
)
from django.contrib.contenttypes.models import ContentType
from typing import Any
from django.views.generic.edit import ModelFormMixin


class HtmxMixin:
    '''Updates the template name based on the request.htmx attribute.'''
    htmx_template = 'bloomerp_htmx_base_view.html'
    base_detail_template = 'detail_views/bloomerp_base_detail_view.html'
    htmx_detail_target = 'detail-content'
    htmx_main_target = 'main-content'
    is_detail_view = False

    def get_context_data(self, **kwargs:Any) -> dict:
        try:
            context = super().get_context_data(**kwargs)
        except AttributeError:
            # If the super class does not have a get_context_data method
            context = {}

        # ---------------------
        # NORMAL REQUEST
        # ---------------------
        if not self.request.htmx or self.request.htmx.history_restore_request:
            if self.is_detail_view or isinstance(self, DetailView) or isinstance(self, UpdateView):
                context['include_main_content'] = self.base_detail_template
                context['include_detail_content'] = self.template_name
                context['template_name'] = self.base_detail_template
            else:
                context['template_name'] = self.template_name
            self.template_name = self.htmx_template
        
        # ---------------------
        # HTMX REQUEST
        # ---------------------
        else:
            # Check the target of htmx
            if self.request.htmx.target == self.htmx_main_target:
                if isinstance(self, DetailView) or isinstance(self, UpdateView):
                    # In this case, we are dealing with a detail view
                    context['include_detail_content'] = self.template_name
                    self.template_name = self.base_detail_template
        return context
    
class BloomerpModelFormViewMixin(ModelFormMixin):
    '''
    A mixin that provides a form view for a model.

    It includes the following features:

        - It uses the BloomerpModelForm form class
        - It sets the user and model attributes on the form
        - It saves the form instance to the database
        - It sets the updated_by attribute on the instance if it exists
        - It sets the created_by attribute on the instance if it exists
    '''
    exclude = []
    form_class = BloomerpModelForm

    def get_form_kwargs(self) -> dict:
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        kwargs["model"] = self.model
        return kwargs
    
    def form_valid(self, form: BloomerpModelForm) -> HttpResponse:
        # Call form valid on super class to make sure messages are displayed
        super().form_valid(form)
        
        # Save the form instance but don't commit to the database yet
        obj = form.save(commit=False)

        # Check if the instance has 'last_updated_by' attribute and set it
        if hasattr(obj, "updated_by"):
            obj.updated_by = self.request.user

        # Check if the instance has 'created_by' attribute and set it
        if hasattr(obj, "created_by") and not obj.created_by:
            obj.created_by = self.request.user

        # Now save the object to the database
        obj.save()

        # Check if the form has a save_m2m method and call it
        if hasattr(form, "save_m2m"):
            form.save_m2m()
        
        # Check if the form has an update_file_fields method and call it
        if hasattr(obj, "save_file_fields"):
            obj.save_file_fields()

        return redirect(self.get_success_url())
    
    def get_form(self, form_class=None) -> BloomerpModelForm:
        form = super().get_form(form_class)

        if "updated_by" in form.fields:
            del form.fields["updated_by"]

        if "created_by" in form.fields:
            del form.fields["created_by"]
        
        return form

    def get_form_class(self) -> BloomerpModelForm:
        return modelform_factory(
            model=self.model,
            fields=self.fields,
            form=BloomerpModelForm,
            exclude=self.exclude
        )

from bloomerp.models import BloomerpModel
class BloomerpModelContextMixin:
    '''
    A mixin that provides context data whenever rendering a model view.
    This mixin provides the following context data:
        - user
        - model_name
        - model_name_plural
        - content_type_id
        - model_dashboard_url
        - create_view_url
        - update_view_url
        - list_view_url
        - detail_view_url

    Note: consider splitting this mixin into list and detail mixins.
    '''
    model: BloomerpModel = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Check if the model attribute is set
        if not self.model:
            raise NotImplementedError("You must provide a model attribute to the view.")

        # init content type
        content_type = ContentType.objects.get_for_model(self.model)

        self.view_content_type = content_type
        
        # Detail view routes
        detail_view_tabs = UserDetailViewTab.get_detail_view_tabs(content_type=content_type, user=self.request.user)
        if not detail_view_tabs:
            detail_view_tabs = UserDetailViewTab.generate_default_for_user(content_type=content_type, user=self.request.user)

        context['detail_view_tabs'] = detail_view_tabs
        context['detail_links_form'] = DetailLinksSelectForm(content_type=content_type, user=self.request.user)


        # User context data
        context["user"] = self.request.user

        # Model context data
        context["model_name"] = self.model._meta.verbose_name
        context["model_name_plural"] = self.model._meta.verbose_name_plural
        context["content_type_id"] = ContentType.objects.get_for_model(self.model).pk
        context["model"] = self.model

        # URL context data
        context["model_dashboard_url"] = get_model_dashboard_view_url(self.model)
        context["create_view_url"] = get_create_view_url(self.model)
        context['update_view_url'] = get_update_view_url(self.model)
        context['list_view_url'] = get_list_view_url(self.model)
        context['detail_view_url'] = get_detail_view_url(self.model)
        context['bulk_upload_url'] = get_bulk_upload_view_url(self.model)

        # Application fields context data
        context['application_fields'] = ApplicationField.objects.filter(content_type_id=context['content_type_id'])
        return context


from django.shortcuts import render
from bloomerp.utils.router import BloomerpRouter
from .models import Employee, Customer, BloomerpModel
from bloomerp.views.mixins import HtmxMixin
from django.views import View
from django.views.generic import TemplateView
from bloomerp.views.core import BloomerpBaseDetailView
import time

router = BloomerpRouter()

# -----------------------------
# LIST VIEW ROUTE EXAMPLE
# -----------------------------
@router.bloomerp_route(
    path='send-emails', # path will become 
    name='Send Emails',
    description='Send email to {model}',
    route_type='list',
    url_name='send_emails', # url name will become employees_send_emails & customers_send_emails
    models=[Employee, Customer] # List of models for which the route will be created
)
class SendEmailsView(HtmxMixin, View):
    template_name = 'email_employees.html'
    model : BloomerpModel = None # Model will be passed through via the router

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        time.sleep(2) # Simulate a slow request
        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data()

        # Send emails business logic
        # ...

        return render(request, self.template_name, context)
        
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        # For example: get objects from model
        context['objects'] = self.model.objects.all()
        context['model'] = self.model
        return context


# -----------------------------
# DETAIL VIEW ROUTE EXAMPLE
# -----------------------------
@router.bloomerp_route(
    path='send-emails', 
    name='Send Email',
    description='Send email to object of {model} model',
    route_type='detail',
    url_name='send_email', 
    models=[Employee, Customer]
)
class SendEmailView(BloomerpBaseDetailView):
    template_name = 'send_email.html'


    def post(self, request, *args, **kwargs):
        context = self.get_context_data()

        # Send email business logic
        # ...

        return render(request, self.template_name, context)


# -----------------------------
# APP LEVEL ROUTE EXAMPLE
# -----------------------------
@router.bloomerp_route(
    path='custom-dashboard', # path will become /custom-dashboard
    name='Custom Dashboard',
    description='Custom dashboard created for the app',
    route_type='app',
    url_name='custom_dashboard', # url name will become custom_dashboard
)
class CustomDashboardView(HtmxMixin, View):
    template_name = 'custom_dashboard.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return render(request, self.template_name, context)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['meaning_of_life'] = 42
        return context
    

from test_app.models import EmployeeOnAccommodation
@router.bloomerp_route(
    path='test/',
    name='test',
    description='test',
    route_type='app',
    url_name='test'
)
class TestView(HtmxMixin, TemplateView):
    template_name = 'test.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context


from django.db.models import Model
import random
import string
from datetime import datetime

def random_object_factory(model: Model):
    '''
    Function that generates a random object for a particular model.
    '''
    try:
        # Get the fields of the model
        fields = model._meta.get_fields()

        # Initialize the data dictionary
        data = {}

        # Loop through the fields
        for field in fields:
            if field.get_internal_type() == 'CharField':
                data[field.name] = ''.join(random.choices(string.ascii_letters, k=10))

            elif field.get_internal_type() == 'IntegerField':
                data[field.name] = random.randint(1, 100)

            elif field.get_internal_type() == 'DateTimeField':
                data[field.name] = datetime.now().isoformat()

            elif field.get_internal_type() == 'ForeignKey':
                if field.one_to_many or field.many_to_many:
                    continue

                related_model = field.related_model
                related_instance = related_model.objects.first()
                if related_instance:
                    data[field.name] = related_instance
                else:
                    raise ValueError(f"No instances found for related model {related_model}")

            elif field.get_internal_type() == 'BooleanField':
                data[field.name] = random.choice([True, False])

            elif field.get_internal_type() == 'FloatField':
                data[field.name] = random.uniform(1.0, 100.0)

            elif field.get_internal_type() == 'TextField':
                data[field.name] = ''.join(random.choices(string.ascii_letters + string.digits, k=50))

            elif field.get_internal_type() == 'EmailField':
                data[field.name] = ''.join(random.choices(string.ascii_letters, k=5)) + '@example.com'

            elif field.get_internal_type() == 'URLField':
                data[field.name] = 'https://www.' + ''.join(random.choices(string.ascii_letters, k=5)) + '.com'

            elif field.get_internal_type() == 'DateField':
                data[field.name] = datetime.now().date().isoformat()
            
            elif field.get_internal_type() == 'DecimalField':
                data[field.name] = random.uniform(1.0, 100.0)

            elif field.get_internal_type() == 'PositiveIntegerField':
                data[field.name] = random.randint(1, 100)

            elif field.get_internal_type() == 'ManyToManyField':
                continue  # Handle ManyToManyField after instance creation

        # Create the instance without saving

        instance = model(**data)
        
        return instance

    except Exception as e:
        print(f"An error occurred with field {field}: {e}")
        return None






# Generated by Django 5.1.1 on 2024-09-26 23:43

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0020_alter_documenttemplate_template_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserDashboardView",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime_created", models.DateTimeField(auto_now_add=True)),
                ("datetime_updated", models.DateTimeField(auto_now=True)),
                (
                    "name",
                    models.CharField(
                        help_text="The name of the dashboard", max_length=255
                    ),
                ),
                ("kpis", models.ManyToManyField(to="bloomerp.kpi")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dashboard_view",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "bloomerp_user_dashboard_view",
                "managed": True,
            },
        ),
        migrations.DeleteModel(
            name="UserKpiListViewPreference",
        ),
    ]



from django.db import models
from django.db.models import Q, Value, F
from django.db.models.functions import Concat
from django.urls import reverse
from django.core.exceptions import ValidationError

class StringSearchModelMixin(models.Model):
    """
    A mixin for models that need to be searchable by a string query.
    """
    string_search_fields: list = None  # The list of fields to search in
    allow_string_search: bool = None

    class Meta:
        abstract = True

    @classmethod
    def string_search(cls, query: str):
        '''
        Static method to search in all string fields of the model.
        Returns a QuerySet filtered by the query in all CharField or TextField attributes.
        '''
        # Get all string fields (CharField and TextField) of the model
        if cls.string_search_fields:
            string_fields = cls.string_search_fields
        else:
            string_fields = [
                field.name for field in cls._meta.fields
                if isinstance(field, models.CharField) or isinstance(field, models.TextField)
            ]

        queryset = cls.objects.all()
        
        # Replace spaces in the query with empty strings

        # Build a Q object to filter across all string fields
        query_filter = Q()
        for field in string_fields:
            if '+' in field:
                concatenated_query = query.replace(' ','')
                concat_fields = field.split('+')
                concat_operation = Concat(*[F(f) if f != ' ' else Value(' ') for f in concat_fields], output_field=models.CharField())
                queryset = queryset.annotate(**{field: concat_operation})
                query_filter |= Q(**{f"{field}__icontains": concatenated_query})
            else:
                query_filter |= Q(**{f"{field}__icontains": query})

        # Filter the queryset by the query in any of the string fields
        return queryset.filter(query_filter)
    
class TimestampedModelMixin(models.Model):
    """
    A mixin for models that need to be timestamped.
    """
    datetime_created = models.DateTimeField(auto_now_add=True)
    datetime_updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

class UserStampedModelMixin(models.Model):
    """
    A mixin for models that need to be stamped with the user that created or updated them.
    """
    created_by = models.ForeignKey('bloomerp.User', on_delete=models.SET_NULL, related_name='%(class)s_created', null=True)
    updated_by = models.ForeignKey('bloomerp.User', on_delete=models.SET_NULL, related_name='%(class)s_updated', null=True)

    class Meta:
        abstract = True

class AbsoluteUrlModelMixin(models.Model):
    """
    A mixin for models that need to have an absolute URL.
    """
    class Meta:
        abstract = True

    def get_absolute_url(self):
        """
        Returns the absolute URL of the model instance.
        """
        return reverse(f'{self._meta.verbose_name_plural.replace(' ','_')}_detail_overview'.lower(), kwargs={'pk': self.pk})
    
class ContentLayoutModelMixin(models.Model):
    """
    A mixin for models that need to have a content layout.
    """
    class Meta:
        abstract = True

    content_layout = models.JSONField(default=dict)

    items_field_name: str = None  # The name of the field that contains the items in the layout, is a many to many field
    max_rows: int = None  # The maximum number of rows in the layout
    max_columns: int = None  # The maximum number of columns in each row
    max_items: int = None  # The maximum number of items in each column

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check if the items_field_name is defined
        if not self.items_field_name:
            raise NotImplementedError(f"{self.__class__.__name__} must define items_field_name.")
        
        # Check if the items_field_name exists in the model and is a many to many field
        items_field = self._meta.get_field(self.items_field_name)
        if not items_field.many_to_many:
            raise ValueError(f"{self.items_field_name} must be a many to many field.")

    def generate_layout(self) -> dict:
        """
        Generates a layout based on the items in the many-to-many field.
        """
        items = getattr(self, self.items_field_name).all()
        layout = {"rows": []}
        row = {"size": 12, "columns": []}
        column = {"size": 12 // self.max_columns, "items": []}

        for item in items:
            if len(column["items"]) >= self.max_items:
                row["columns"].append(column)
                column = {"size": 12 // self.max_columns, "items": []}
                if len(row["columns"]) >= self.max_columns:
                    layout["rows"].append(row)
                    row = {"size": 12, "columns": []}
                    if len(layout["rows"]) >= self.max_rows:
                        break
            column["items"].append(item.pk)

        if column["items"]:
            row["columns"].append(column)
        if row["columns"]:
            layout["rows"].append(row)

        return layout

    def clean(self):
        """
        Validates the content layout.
        """
        layout: dict = self.content_layout
        errors = []

        # Check if layout is a dictionary
        if not isinstance(layout, dict):
            errors.append(ValidationError("Content layout must be a dictionary.", code='invalid'))

        # Check if layout contains 'rows' key
        if 'rows' not in layout:
            errors.append(ValidationError("Content layout must contain 'rows' key.", code='missing_rows'))

        # Check if the row count is within the limits
        if 'rows' in layout and len(layout.get('rows')) > self.max_rows:
            errors.append(ValidationError(f"Content layout exceeds the maximum number of rows ({self.max_rows}).", code='max_rows_exceeded'))

        # Get the set of valid item IDs from the many-to-many field
        items_field = self._meta.get_field(self.items_field_name)
        valid_item_ids = set(items_field.related_model.objects.values_list('id', flat=True))

        # Check if the column count is within the limits
        if 'rows' in layout:
            for row in layout.get('rows'):
                if len(row.get('columns')) > self.max_columns:
                    errors.append(ValidationError(f"Row exceeds the maximum number of columns ({self.max_columns}).", code='max_columns_exceeded'))

                # Check if the item count is within the limits
                for column in row.get('columns'):
                    if len(column.get('items')) > self.max_items:
                        errors.append(ValidationError(f"Column exceeds the maximum number of items ({self.max_items}).", code='max_items_exceeded'))

                    # Check if the item IDs are valid
                    for item in column.get('items'):
                        if item not in valid_item_ids:
                            errors.append(ValidationError(f"Invalid item ID {item} in content layout.", code='invalid_item_id'))

        # Raise all collected errors at once
        if errors:
            raise ValidationError({'content_layout': errors})

        # Call the clean method of the superclass
        super().clean()

    def save(self, *args, **kwargs):
        """
        Overrides the save method to generate the layout if it's empty or needs to be updated.
        """
        super().save(*args, **kwargs)
        if not self.content_layout:
            self.content_layout = self.generate_layout()
            super().save(update_fields=['content_layout'])


class AvatarModelMixin(models.Model):
    """
    A mixin for models that need to have an avatar.
    """
    class Meta:
        abstract = True

    avatar = models.ImageField(upload_to='avatars/', null=True, blank=True)


# Generated by Django 5.1.1 on 2024-11-11 23:43

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0046_workspace_content_type_workspace_user"),
    ]

    operations = [
        migrations.RenameField(
            model_name="comment",
            old_name="comment",
            new_name="content",
        ),
        migrations.RemoveField(
            model_name="comment",
            name="user",
        ),
    ]



from bloomerp.models.auth import User
from django.db.models import Model
from bloomerp.utils.models import model_name_singular_underline


def can_view_model(user:User, model:Model) -> bool:
    """
    This function returns True if the user has permission to view the model.
    """
    return user.has_perm('view_%s' % model_name_singular_underline(model))

def can_add_model(user:User, model:Model) -> bool:
    """
    This function returns True if the user has permission to add the model.
    """
    return user.has_perm('add_%s' % model_name_singular_underline(model))

def can_change_model(user:User, model:Model) -> bool:
    """
    This function returns True if the user has permission to change the model.
    """
    return user.has_perm('change_%s' % model_name_singular_underline(model))

def can_delete_model(user:User, model:Model) -> bool:
    """
    This function returns True if the user has permission to delete the model.
    """
    return user.has_perm('delete_%s' % model_name_singular_underline(model))


# Generated by Django 5.1.1 on 2024-09-17 19:31

import bloomerp.models
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0012_alter_file_file"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="applicationfield",
            name="related_model",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="related_models",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="file",
            name="file",
            field=models.FileField(upload_to=bloomerp.models.File.upload_to),
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from django.contrib.contenttypes.models import ContentType
from django.forms.models import modelform_factory
from bloomerp.forms.core import BloomerpModelForm
from bloomerp.models import User
from django.contrib.auth.decorators import login_required
from django import forms
from bloomerp.utils.requests import parse_bool_parameter

@login_required
@route('object_model_form')
def object_model_form(request:HttpRequest) -> HttpResponse:
    '''
    Component to create an object of a model.
    Requires content_type_id and form_prefix in the query parameters.
    '''
    # Get the content type id and form prefix from the query parameters
    content_type_id = request.GET.get('content_type_id')
    form_prefix = request.GET.get('form_prefix','')
    object_id = request.GET.get('object_id', None)
    fields = request.GET.getlist('field_name', None)
    fields_hidden = request.GET.getlist('field_name_hidden', None)
    reset_on_submit = request.GET.get('reset_on_submit', False)
    hide_default_fields = parse_bool_parameter(request.GET.get('hide_default_fields'), True)
    apply_helper = parse_bool_parameter(request.GET.get('apply_helper'), True)


    # Add permission check here
    user : User = request.user

    # Attributes to be passed to the template
    created = False
    new_object = None


    if not content_type_id:
        return HttpResponse('content_type_id required in the query parameters', status=400)
    
    # Get the model and create a form
    model = ContentType.objects.get(id=content_type_id).model_class()

    # Permission check
    if not user.has_perm(f'{model._meta.app_label}.add_{model._meta.model_name}') and not user.has_perm(f'{model._meta.app_label}.change_{model._meta.model_name}'):
        return HttpResponse('User does not have permission to add objects of this model')

    # Create the form
    if fields:
        Form : BloomerpModelForm = modelform_factory(model, fields=fields, form=BloomerpModelForm)
    else:
        Form : BloomerpModelForm = modelform_factory(model, fields='__all__', form=BloomerpModelForm)

    if request.method == 'POST':
        form : BloomerpModelForm = Form(
            data=request.POST, 
            files=request.FILES, 
            prefix=form_prefix, 
            model=model, 
            hide_default_fields=hide_default_fields,
            apply_helper=apply_helper
            )
        if form.is_valid():
            form.save()
            created = True
            new_object = form.instance

            if hasattr(form, 'save_m2m'):
                form.save_m2m()

        if reset_on_submit:
            form = Form(prefix=form_prefix, model=model, user=user, hide_default_fields=hide_default_fields, apply_helper=apply_helper)
    else:
        if fields_hidden:
            Form = hide_fields(Form, fields_hidden)
        
        if object_id:
            instance = model.objects.get(id=object_id)
            form = Form(instance=instance, prefix=form_prefix, model=model, user=user, hide_default_fields=hide_default_fields, apply_helper=apply_helper)
        else:
            form = Form(prefix=form_prefix, model=model, user=user, hide_default_fields=hide_default_fields, apply_helper=apply_helper)

    return render(request, 'components/object_model_form.html', {'form': form, 'created': created, 'form_prefix': form_prefix, 'new_object': new_object, 'reset_on_submit': reset_on_submit})


def hide_fields(Form, fields) -> BloomerpModelForm:
    '''
    Hides the fields in the form
    '''
    for field in fields:
        try:
            Form.base_fields[field].widget = forms.HiddenInput()
        except KeyError:
            pass
    return Form

